package edu.georgetown.gucs.experiment;

import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.sql.Blob;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.StringTokenizer;
import java.util.Vector;

import org.postgresql.largeobject.LargeObject;
import org.postgresql.largeobject.LargeObjectManager;

import edu.georgetown.gucs.dictionary.Dictionary;
import edu.georgetown.gucs.utility.ExceptionHandler;
import edu.georgetown.gucs.utility.Global;

/**
 * Creates a database of the experiment results
 * 
 * @author Clay Shields
 * @author Lindsay Neubauer
 * 
 */
public class DBInterface {

  private Connection db;
  private Statement sql;
  private ResultSet results;

  // assume system user name is same as DB role name
  private String username = "testuser";

  // "If no password is specified, the password will be set to null and
  // password authentication will always fail for that user." So, we should
  // create our roles with a non-null password.
  // http://www.postgresql.org/docs/9.0/static/sql-createrole.html
  private String password = "testPassWord";

  private int filesAddedCount = 0;
  private int fileMatchesAddedCount = 0;
  private int fingerprintMatchAddedCount = 0;
  private int batchSize = 100000;

  private Map<String, Integer> fileCache;
  private Map<Integer, HashMap<Integer, Integer>> fileMatchCache;
  private Map<Integer, HashMap<Integer, Integer>> fingerprintMatchCache;

  private PreparedStatement addFileMatch;
  private PreparedStatement insertFile;
  private PreparedStatement insertFingerprintMatch;

  /**
   * Constructor that connects to a database
   * 
   * @param databaseName
   *          the string name of the database to connect to
   */
  public DBInterface(String databaseName) {
    this.fileCache = new HashMap<String, Integer>();
    this.fileMatchCache = new HashMap<Integer, HashMap<Integer, Integer>>();
    this.fingerprintMatchCache = new HashMap<Integer, HashMap<Integer, Integer>>();

    try {
      // load the driver and connect to the database
      Class.forName("org.postgresql.Driver");
      this.db = DriverManager.getConnection("jdbc:postgresql:" + databaseName, this.username, this.password);
      this.sql = this.db.createStatement();

      this.addFileMatch = this.db
          .prepareStatement("insert into file_match (file_id, match_id, IDf_lower_bound, IDf_upper_bound) values (?,?,?,?)");
      this.insertFile = this.db.prepareStatement("insert into file (full_path) values (?)");
      this.insertFingerprintMatch = this.db
          .prepareStatement("insert into fingerprint_match (trial_id, source_print_id, "
              + "target_print_id, does_match, should_match, source_mangler_id, target_mangler_id, match_score) "
              + "values (?,?,?,?,?,(select mangler_id from fingerprint where fingerprint_id=?),"
              + "(select mangler_id from fingerprint where fingerprint_id = ?),?)");

    } catch (SQLException e) {
      ExceptionHandler.handleSQLException(e);
    } catch (ClassNotFoundException e) {
      String exceptionMessage = String.format("%s%n%s%n%s%n%s%n%s%n%s%n%n%s%n",
          "An error here could indicate that the database for the dataset does not exist yet.",
          "Create a new database by running:", "psql", "and then typing:",
          "create database " + databaseName, "Then enter that database and load the database layout.",
          e.getMessage());
      ExceptionHandler.handleExceptionThrow(new ClassNotFoundException(exceptionMessage));
    }
  }

  /**
   * Starts a new experiment run
   * 
   * @return the <code>int</code> new run number
   */
  public synchronized int startNewRun() {
    PreparedStatement startRun = null;
    try {
      startRun = this.db.prepareStatement("insert into run (start_time) (select localtimestamp)");
      startRun.executeUpdate();
      startRun.close();
    } catch (SQLException e) {
      ExceptionHandler.handleSQLException(e, this.db, startRun);
    }

    PreparedStatement runNumber = null;
    int currentValue = 0;
    try {
      runNumber = this.db.prepareStatement("select currval('run_run_number_seq')");
      this.results = runNumber.executeQuery();
      if (this.results.next()) {
        currentValue = this.results.getInt("currval");
      }
      runNumber.close();
    } catch (SQLException e) {
      ExceptionHandler.handleSQLException(e, this.db, runNumber);
    }
    return currentValue;
  }

  /**
   * Finishes an experiment run
   * 
   * @param runNumber
   *          the <code>int</code> run number to finish
   */
  public void finishRun(int runNumber) {
    PreparedStatement endRun = null;
    try {
      endRun = this.db.prepareStatement("update run set end_time = (select localtimestamp) where run_number=?");
      endRun.setInt(1, runNumber);
      endRun.executeUpdate();
      endRun.close();
    } catch (SQLException e) {
      ExceptionHandler.handleSQLException(e, this.db, endRun);
    }
  }

  /**
   * Removes the experiment results from experiment_results table
   * 
   * @param experimentID
   *          the <code>int</code> experiment ID to remove
   */
  public void clearExperimentResults(int experimentID) {
    PreparedStatement clearExpResults = null;
    try {
      clearExpResults = this.db.prepareStatement("delete from experiment_results where experiment_id = ?");
      clearExpResults.setInt(1, experimentID);
      clearExpResults.executeUpdate();
      clearExpResults.close();
    } catch (SQLException e) {
      ExceptionHandler.handleSQLException(e, this.db, clearExpResults);
    }
  }

  /**
   * Inserts experiment results into database; computes average, standard deviation and error for precision, recall,
   * fscore and dictionary size
   * 
   * @param experimentID
   *          the <code>int</code> experiment ID for results to insert
   * @param numTrials
   *          the <code>int</code> number of trials run
   */
  public void insertExperimentResults(int experimentID, int numTrials) {
    PreparedStatement calcExperimentResults = null;
    PreparedStatement experimentAverageDictionarySize = null;
    PreparedStatement experimentStdevDictionarySize = null;
    PreparedStatement experimentErrDictionarySize = null;
    PreparedStatement experimentErrPrecision = null;
    PreparedStatement experimentErrRecall = null;
    PreparedStatement experimentErrFscore = null;
    final double tValue = Statistics.getTValueNinetyFive(numTrials - 1);

    try {
      calcExperimentResults = this.db
          .prepareStatement("insert into experiment_results (experiment_id, "
              + "print_mangler_id, average_precision, average_recall, average_fscore, stdev_precision, stdev_recall, stdev_fscore) "
              + "select ?,print_mangler_id,avg(precision),avg(recall),avg(fscore),stddev(precision),stddev(recall),stddev(fscore)"
              + "from trial_results where trial_id in (select trial_id from experiment_trial where experiment_id=?) "
              + "group by print_mangler_id order by print_mangler_id");
      calcExperimentResults.setInt(1, experimentID);
      calcExperimentResults.setInt(2, experimentID);
      calcExperimentResults.executeUpdate();
      calcExperimentResults.close();
    } catch (SQLException e) {
      ExceptionHandler.handleSQLException(e, this.db, calcExperimentResults);
    }

    try {
      experimentAverageDictionarySize = this.db.prepareStatement("update experiment_results "
          + "set average_dictionary_size = (select avg(size) from dictionary where dictionary_id in "
          + "(select dictionary_id from trial where trial_id in "
          + "(select trial_id from experiment_trial where experiment_id = ?))) where experiment_id = ? ");
      experimentAverageDictionarySize.setInt(1, experimentID);
      experimentAverageDictionarySize.setInt(2, experimentID);
      experimentAverageDictionarySize.executeUpdate();
      experimentAverageDictionarySize.close();
    } catch (SQLException e) {
      ExceptionHandler.handleSQLException(e, this.db, experimentAverageDictionarySize);
    }

    try {
      experimentStdevDictionarySize = this.db
          .prepareStatement("update experiment_results "
              + "set stdev_dictionary_size = (select stddev(size) from dictionary where dictionary_id in "
              + "(select dictionary_id from trial where trial_id in (select trial_id from experiment_trial where experiment_id = ?))) "
              + "where experiment_id = ? ");
      experimentStdevDictionarySize.setInt(1, experimentID);
      experimentStdevDictionarySize.setInt(2, experimentID);
      experimentStdevDictionarySize.executeUpdate();
      experimentStdevDictionarySize.close();
    } catch (SQLException e) {
      ExceptionHandler.handleSQLException(e, this.db, experimentStdevDictionarySize);
    }

    try {
      experimentErrDictionarySize = this.db.prepareStatement("update experiment_results "
          + "set err_95_bound_dictionary_size = (stdev_dictionary_size/sqrt(?))*?;");
      experimentErrDictionarySize.setInt(1, numTrials);
      experimentErrDictionarySize.setDouble(2, tValue);
      experimentErrDictionarySize.executeUpdate();
      experimentErrDictionarySize.close();
    } catch (SQLException e) {
      ExceptionHandler.handleSQLException(e, this.db, experimentErrDictionarySize);
    }

    try {
      experimentErrPrecision = this.db.prepareStatement("update experiment_results "
          + "set err_95_bound_precision = (stdev_precision/sqrt(?))*?;");
      experimentErrPrecision.setInt(1, numTrials);
      experimentErrPrecision.setDouble(2, tValue);
      experimentErrPrecision.executeUpdate();
      experimentErrPrecision.close();
    } catch (SQLException e) {
      ExceptionHandler.handleSQLException(e, this.db, experimentErrPrecision);
    }

    try {
      experimentErrRecall = this.db.prepareStatement("update experiment_results "
          + "set err_95_bound_recall = (stdev_recall/sqrt(?))*?;");
      experimentErrRecall.setInt(1, numTrials);
      experimentErrRecall.setDouble(2, tValue);
      experimentErrRecall.executeUpdate();
      experimentErrRecall.close();
    } catch (SQLException e) {
      ExceptionHandler.handleSQLException(e, this.db, experimentErrRecall);
    }

    try {
      experimentErrFscore = this.db.prepareStatement("update experiment_results "
          + "set err_95_bound_fscore = (stdev_fscore/sqrt(?))*?;");
      experimentErrFscore.setInt(1, numTrials);
      experimentErrFscore.setDouble(2, tValue);
      experimentErrFscore.executeUpdate();
      experimentErrFscore.close();
    } catch (SQLException e) {
      ExceptionHandler.handleSQLException(e, this.db, experimentErrFscore);
    }
  }

  /**
   * Closes this database
   */
  public void close() {
    try {
      this.sql.close();
      this.db.close();
    } catch (SQLException e) {
      ExceptionHandler.handleSQLException(e);
    }
  }

  /**
   * Closes this database
   */
  @Override
  protected void finalize() {
    close();
  }

  /**
   * Checks whether the file exists in the database
   * 
   * @param file
   *          the file to check
   * 
   * @return the <code>boolean</code> whether the file exists in the database
   */
  public boolean fileExists(File file) {
    return (getFileID(file) > 0);
  }

  /**
   * Provides the ID of the file or 0 if it is not in the this database
   * 
   * @param file
   *          the file to check
   * 
   * @return the <code>int</code> file ID; returns 0 if file does not exist in database
   */
  public int getFileID(File file) {
    int fileID = 0;
    if (this.fileCache.containsKey(file.getAbsolutePath())) {
      return this.fileCache.get(file.getAbsolutePath());
    }
    PreparedStatement fileIDFromFile = null;
    try {
      fileIDFromFile = this.db.prepareStatement("select file_id from file where full_path=?");
      fileIDFromFile.setString(1, file.getAbsolutePath());
      this.results = fileIDFromFile.executeQuery();

      if (this.results.next()) {
        fileID = this.results.getInt("file_id");
        if (fileID != 0) {
          this.fileCache.put(file.getAbsolutePath(), fileID);
        }
      }
      fileIDFromFile.close();
    } catch (SQLException e) {
      ExceptionHandler.handleSQLException(e, this.db, fileIDFromFile);
    }
    return 0;
  }

  /**
   * Adds a file and its match to fileMatchCache
   * 
   * @param fileID
   *          the <code>int</code> file ID
   * @param matchID
   *          the <code>int</code> match ID
   */
  private void addToFileMatchCache(int fileID, int matchID) {
    HashMap<Integer, Integer> file_match;
    if ((file_match = this.fileMatchCache.get(fileID)) != null) {
      if (!file_match.containsKey(matchID)) {
        file_match.put(matchID, null);
      }
    } else {
      file_match = new HashMap<Integer, Integer>();
      file_match.put(matchID, null);
      this.fileMatchCache.put(fileID, file_match);
    }
  }

  /**
   * Checks if file and its match are in fileMatchCache
   * 
   * @param fileID
   *          the <code>int</code> file ID
   * @param matchID
   *          the <code>int</code> match ID
   * 
   * @return the <code>boolean</code> whether the file and match are in the fileMatchCache
   */
  private boolean isFileMatchInCache(int fileID, int matchID) {
    if (this.fileMatchCache == null || !this.fileMatchCache.containsKey(fileID)) {
      return false;
    } else if (this.fileMatchCache.get(fileID).containsKey(matchID)) {
      return true;
    }
    return false;
  }

  /**
   * Loads files and matches into the fileMatchCache
   */
  private void loadFileMatchCache() {
    int fileID, pathID;
    try {
      this.results = this.sql.executeQuery("select file_id, match_id from file_match");
      while (this.results.next()) {
        fileID = this.results.getInt(1);
        pathID = this.results.getInt(2);
        addToFileMatchCache(fileID, pathID);
      }
      // Try and recover some memory
      this.results = null;
    } catch (SQLException e) {
      ExceptionHandler.handleSQLException(e, this.db, this.sql);
    }
  }

  /**
   * Loads all the existing matches into a cache of file matches
   */
  public void startBulkFileMatchAdd() {
    loadFileMatchCache();
    try {
      this.sql.executeUpdate("drop index file_match_file_id");
      this.sql.executeUpdate("drop index file_match_match_id");
    } catch (SQLException e) {
      // SQL exception means that the index isn't built so it can't be dropped
    }
    this.fileMatchesAddedCount = 0;
  }

  /**
   * Adds file matches and their normalized min and max IDF to the file_match table
   * 
   * @param file
   *          the matching file
   * @param match
   *          the matching file
   * @param minIDF
   *          the <code>double</code> minimum normalized IDF
   * @param maxIDF
   *          the <code>double</code> maximum normalized IDF
   */
  public void bulkFileMatchAdd(File file, File match, double minIDF, double maxIDF) {
    int fileID = getFileID(file);
    int matchID = getFileID(match);

    if (!isFileMatchInCache(fileID, matchID)) {
      try {
        this.addFileMatch.setInt(1, fileID);
        this.addFileMatch.setInt(2, matchID);
        this.addFileMatch.setDouble(3, minIDF);
        this.addFileMatch.setDouble(4, maxIDF);
        this.addFileMatch.addBatch();
        addToFileMatchCache(fileID, matchID);
        this.fileMatchesAddedCount++;

        if ((this.fileMatchesAddedCount % this.batchSize) == 0) {
          this.addFileMatch.executeBatch();
        }
      } catch (SQLException e) {
        ExceptionHandler.handleSQLException(e, this.db, this.addFileMatch);
      }
    }
  }

  /**
   * Rebuild indexes for file_match table
   */
  public void finishBulkFileMatchAdd() {
    try {
      if (this.fileMatchesAddedCount > 0) {
        Global.monitor.printMessage("DBInterface", "Added this many file matches: " + this.fileMatchesAddedCount);
        this.addFileMatch.executeBatch();
      }
    } catch (SQLException e) {
      ExceptionHandler.handleSQLException(e, this.db, this.addFileMatch);
    }

    try {
      this.sql.executeUpdate("create index file_match_file_id on file_match using hash (file_id)");
      this.sql.executeUpdate("create index file_match_match_id on file_match using hash (match_id)");
    } catch (SQLException e) {
      ExceptionHandler.handleSQLException(e, this.db, this.sql);
    }
  }

  /**
   * Checks if files match and if so adds files and their normalized min and max IDF to the file_match table
   * 
   * @param source
   *          the file to check for match
   * @param target
   *          the file to check for match
   * @param minIDF
   *          the <code>double</code> minimum normalized IDF
   * @param maxIDF
   *          the <code>double</code> maximum normalized IDF
   */
  public void addFileMatch(File source, File target, double minIDF, double maxIDF) {
    if (!doFilesMatch(source, target)) {
      int sourceID = addFile(source);
      int targetID = addFile(target);
      try {
        this.addFileMatch.setInt(1, sourceID);
        this.addFileMatch.setInt(2, targetID);
        this.addFileMatch.setDouble(3, minIDF);
        this.addFileMatch.setDouble(4, maxIDF);
        this.addFileMatch.addBatch();
      } catch (SQLException e) {
        ExceptionHandler.handleSQLException(e, this.db, this.addFileMatch);
      }
    }
  }

  /**
   * Adds file matches to database
   */
  public void executeFileMatch() {
    try {
      this.addFileMatch.executeBatch();
    } catch (SQLException e) {
      ExceptionHandler.handleSQLException(e, this.db, this.addFileMatch);
    }
  }

  /**
   * Check if files match
   * 
   * @param source
   *          the file to check for a match
   * @param target
   *          the file to check for a match
   * 
   * @return the <code>boolean</code> whether the files match
   */
  private boolean doFilesMatch(File source, File target) {
    int matchID = 0;
    int sourceID = getFileID(source);
    int targetID = getFileID(target);
    PreparedStatement findFileMatch = null;
    try {
      findFileMatch = this.db
          .prepareStatement("select file_match_id from file_match where file_id =(?) and match_id = (?)");
      findFileMatch.setInt(1, sourceID);
      findFileMatch.setInt(2, targetID);
      this.results = findFileMatch.executeQuery();

      if (this.results.next()) {
        matchID = this.results.getInt("file_match_id");
      }
      findFileMatch.close();
    } catch (SQLException e) {
      ExceptionHandler.handleSQLException(e, this.db, findFileMatch);
    }
    return (matchID > 0);
  }

  /**
   * Loads cache of file IDs and their paths
   */
  private void loadFileCache() {
    int fileID;
    String filePath;
    try {
      this.results = this.sql.executeQuery("select file_id, full_path from file");
      while (this.results.next()) {
        fileID = this.results.getInt(1);
        filePath = this.results.getString(2);
        if (!this.fileCache.containsKey(filePath)) {
          this.fileCache.put(filePath, fileID);
        }
      }
      this.results = null;
    } catch (SQLException e) {
      ExceptionHandler.handleSQLException(e, this.db, this.sql);
    }
  }

  /**
   * Drops all the indexes on the file table of this database
   */
  public void startBulkFileAdd() {
    loadFileCache();
    try {
      this.sql.executeUpdate("drop index file_full_path");
      this.filesAddedCount = 0;
    } catch (SQLException e) {
      // An SQL exception here just means that the index isn't built, and
      // therefore
      // can't be dropped, so ignore it.
    }
  }

  /**
   * Adds a file to the file table of this database
   * 
   * @param file
   *          the file to add to this database
   */
  public void bulkFileAdd(File file) {
    try {
      if (!this.fileCache.containsKey(file.getAbsolutePath())) {
        this.insertFile.setString(1, file.getAbsolutePath());
        this.insertFile.addBatch();
        this.filesAddedCount++;

        if ((this.filesAddedCount % this.batchSize) == 0) {
          this.insertFile.executeBatch();
        }
      }
    } catch (SQLException e) {
      ExceptionHandler.handleSQLException(e, this.db, this.insertFile);
    }
  }

  /**
   * Rebuild indexes for file_full_path table
   */
  public void finishBulkFileAdd() {
    try {
      if (this.filesAddedCount > 0) {
        Global.monitor.printMessage("DBInterface", "Added this many files: " + this.filesAddedCount);
        this.insertFile.executeBatch();
      } else {
        Global.monitor.printMessage("DBInterface", "Added no new files");
      }
      // Reload the file index
      this.sql.executeUpdate("create index file_full_path on file using hash (full_path)");
    } catch (SQLException e) {
      ExceptionHandler.handleSQLException(e, this.db, this.insertFile);
    }
  }

  /**
   * Provides the ID for the file if it already exists in the file this database or creates a new ID and adds it
   * 
   * @param file
   *          the file to add to the database
   * 
   * @return the <code>int</code> file ID
   */
  public synchronized int addFile(File file) {
    // Other things depend on being able to add files even if they are there
    int returnValue;

    // File seen is a hash map that caches things we have added or pulled from
    // the database.
    // If the database cache is good, this won't help much
    if (this.fileCache.containsKey(file.getAbsolutePath())) {
      returnValue = this.fileCache.get(file.getAbsolutePath());
    } else {
      returnValue = getFileID(file);
      if (returnValue != 0) {
        this.fileCache.put(file.getAbsolutePath(), returnValue);
      }
    }

    if (0 == returnValue) {
      try {
        this.insertFile.setString(1, file.getAbsolutePath());
        this.insertFile.executeUpdate();
      } catch (SQLException e) {
        ExceptionHandler.handleSQLException(e, this.db, this.insertFile);
      }

      PreparedStatement lastFileID = null;
      try {
        lastFileID = this.db.prepareStatement("select currval('file_file_id_seq')");
        this.results = lastFileID.executeQuery();
        if (this.results.next()) {
          returnValue = this.results.getInt("currval");
          this.fileCache.put(file.getAbsolutePath(), returnValue);
        }
        lastFileID.close();
      } catch (SQLException e) {
        ExceptionHandler.handleSQLException(e, this.db, lastFileID);
      }
    }
    return returnValue;
  }

  /**
   * Matches the given file against the files in this database
   * 
   * @param file
   *          the file to match against all other files in this database
   * 
   * @return the list of files; returns null if the file does not exist is this database
   */
  public List<File> matches(File file) {
    if (0 == getFileID(file)) {
      return null;
    }
    List<File> matchesVec = new Vector<File>();
    PreparedStatement findAllFileMatches = null;
    try {
      findAllFileMatches = this.db
          .prepareStatement("select full_path from file where file_id in (select match_id from file_match where file_id = (select file_id from file where full_path=(?)");
      findAllFileMatches.setString(1, file.getAbsolutePath());
      this.results = null;
      this.results = findAllFileMatches.executeQuery();
      while (this.results.next()) {
        String path = this.results.getString("full_path");
        matchesVec.add(new File(path));
      }
      this.results = null;
      findAllFileMatches.close();
    } catch (SQLException e) {
      ExceptionHandler.handleSQLException(e, this.db, findAllFileMatches);
    }
    return matchesVec;
  }

  /**
   * Updates this database with the random seed to use for this trial
   * 
   * @param trialID
   *          the <code>int</code> trial ID
   * @param randomSeed
   *          the <code>long</code> random seed
   */
  public void changeRandomSeed(int trialID, long randomSeed) {
    PreparedStatement updateRandomSeed = null;
    try {
      updateRandomSeed = this.db.prepareStatement("update trial set random_seed=(?) where trial_id = (?)");
      updateRandomSeed.setLong(1, randomSeed);
      updateRandomSeed.setInt(2, trialID);
      updateRandomSeed.executeUpdate();
      updateRandomSeed.close();
    } catch (SQLException e) {
      ExceptionHandler.handleSQLException(e, this.db, updateRandomSeed);
    }
  }

  /**
   * Adds this dictionary to this trial in this database
   * 
   * @param trialID
   *          the <code>int</code> trial ID
   * @param dictionaryID
   *          the <code>int</code> dictionary ID
   */
  public void addDictionaryToTrial(int trialID, int dictionaryID) {
    PreparedStatement addDictionaryEntryToTrial = null;
    try {
      addDictionaryEntryToTrial = this.db.prepareStatement("update trial set dictionary_id=(?) where trial_id=(?)");
      addDictionaryEntryToTrial.setInt(1, dictionaryID);
      addDictionaryEntryToTrial.setInt(2, trialID);
      addDictionaryEntryToTrial.executeUpdate();
      addDictionaryEntryToTrial.close();
    } catch (SQLException e) {
      ExceptionHandler.handleSQLException(e, this.db, addDictionaryEntryToTrial);
    }
  }

  /**
   * Inserts the compressed dictionary into blob space and stores the size and OID for it in the database
   * 
   * @param dictionary
   *          the trial dictionary to be inserted in this database
   * 
   * @return the <code>int</code> dictionary ID
   */
  public synchronized int newDictionary(Dictionary dictionary) {
    int dictionaryID = 0;

    try {
      // All LargeObject API calls must be within a transaction
      this.db.setAutoCommit(false);

      // Check out code in Blob.java, which was the test file for this.
      LargeObjectManager lom = ((org.postgresql.PGConnection) this.db).getLargeObjectAPI();
      long oid = lom.createLO();
      LargeObject object = lom.open(oid);
      OutputStream out = object.getOutputStream();

      ObjectOutputStream obOut = new ObjectOutputStream(out);
      obOut.writeObject(dictionary);
      obOut.close();
      object.close();

      this.db.commit();
      this.db.setAutoCommit(true);

      PreparedStatement newDictionaryEntry = null;
      try {
        newDictionaryEntry = this.db.prepareStatement("insert into dictionary (size,stored_dictionary) values (?,?)");
        newDictionaryEntry.setInt(1, dictionary.size());
        newDictionaryEntry.setLong(2, oid);
        newDictionaryEntry.executeUpdate();
        newDictionaryEntry.close();
      } catch (SQLException e) {
        ExceptionHandler.handleSQLException(e, this.db, newDictionaryEntry);
      }

      PreparedStatement lastDictionaryID = null;
      try {
        lastDictionaryID = this.db.prepareStatement("select currval('dictionary_dictionary_id_seq')");
        this.results = lastDictionaryID.executeQuery();
        if (this.results.next()) {
          dictionaryID = this.results.getInt("currval");
        }
        lastDictionaryID.close();
      } catch (SQLException e) {
        ExceptionHandler.handleSQLException(e, this.db, lastDictionaryID);
      }

    } catch (IOException e) {
      ExceptionHandler.handleExceptionExit(e);
    } catch (SQLException e) {
      ExceptionHandler.handleSQLException(e);
    }
    return dictionaryID;
  }

  /**
   * Creates a new experiment
   * 
   * @param runNumber
   *          the <code>int</code> run number for this experiment; 0 if not inserting in database
   * @param description
   *          the string human readable description of this experiment
   * @param dataset
   *          the string name of the database to insert this experiment
   * @param datasetPath
   *          the path to the files to use for the dictionary and sample in this experiment
   * @param configFileName
   *          the string filename of the configuration file for this experiment
   * @param fingerprinterName
   *          the string name of the fingerprinter used in this experiment
   * @param matcherName
   *          the string name of the matcher used for fingerprint comparisons in this experiment
   * @param matcherScore
   *          the <code>int</code> minimum score that the matcher considered a match
   * @param minIDF
   *          the <code>double</code> minimum normalized IDF kept in the dictionary
   * @param maxIDF
   *          the <code>double</code> maximum normalized IDF kept in the dictionary
   * @param sampleParameter
   *          the <code>double</code> containing the counts or percents used for the file sample size in this
   *          experiment; numbers less than 1 are processed as percents, numbers greater than or equal to 1 are
   *          processed as counts
   * @param dictionaryParameter
   *          the <code>double</code> containing the counts or percents used for dictionary creation in this experiment;
   *          numbers less than 1 are processed as percents, numbers greater than or equal to 1 are processed as counts
   * @param language
   *          the string language of the files used in this experiment
   * 
   * @return the <code>int<code> experiment ID; returns 0 if experiment does not exist in database
   */
  public synchronized int newExperiment(int runNumber, String description, String dataset, String datasetPath,
      String configFileName, String fingerprinterName, String matcherName, int matcherScore, double minIDF,
      double maxIDF, double sampleParameter, double dictionaryParameter, String language) {

    // TODO: !!! need to add splitter stuff to DBInterface
    
    FileInputStream fis = null;
    int currentValue = 0;
    try {
      // Insert the configuration file as an OID object
      this.db.setAutoCommit(false);
      fis = new FileInputStream(configFileName);

      LargeObjectManager lom = ((org.postgresql.PGConnection) this.db).getLargeObjectAPI();
      long oid = lom.createLO();
      LargeObject object = lom.open(oid);
      OutputStream out = object.getOutputStream();
      int c;
      while ((c = fis.read()) != -1) {
        out.write(c);
      }
      object.close();
      out.close();
      fis.close();
      this.db.commit();
      this.db.setAutoCommit(true);

      PreparedStatement createNewExperiment = null;
      try {
        createNewExperiment = this.db.prepareStatement("insert into experiment (run_number,description,"
            + "dataset,dataset_path,config_file,fingerprinter_name,matcher_name,matcher_parameter,minimum_idf,"
            + "maximum_idf,sample_setting,dictionary_setting,language) values (?,?,?,?,?,?,?,?,?,?,?,?,?)");
        createNewExperiment.setInt(1, runNumber);
        createNewExperiment.setString(2, description);
        createNewExperiment.setString(3, dataset);
        createNewExperiment.setString(4, datasetPath);
        createNewExperiment.setLong(5, oid);
        createNewExperiment.setString(6, fingerprinterName);
        createNewExperiment.setString(7, matcherName);
        createNewExperiment.setInt(8, matcherScore);
        createNewExperiment.setDouble(9, minIDF);
        createNewExperiment.setDouble(10, maxIDF);
        createNewExperiment.setDouble(11, sampleParameter);
        createNewExperiment.setDouble(12, dictionaryParameter);
        createNewExperiment.setString(13, language);
        createNewExperiment.executeUpdate();
        createNewExperiment.close();
      } catch (SQLException e) {
        ExceptionHandler.handleSQLException(e, this.db, createNewExperiment);
      }

      PreparedStatement lastExperimentID = null;
      try {
        lastExperimentID = this.db.prepareStatement("select currval('experiment_experiment_id_seq')");
        this.results = lastExperimentID.executeQuery();

        if (this.results.next()) {
          currentValue = this.results.getInt("currval");
        }
        lastExperimentID.close();
      } catch (SQLException e) {
        ExceptionHandler.handleSQLException(e, this.db, lastExperimentID);
      }

    } catch (SQLException e) {
      ExceptionHandler.handleSQLException(e);
    } catch (FileNotFoundException e) {
      ExceptionHandler.handleExceptionExit(e);
    } catch (IOException e) {
      ExceptionHandler.handleExceptionExit(e);
    } finally {
      try {
        if (fis != null)
          fis.close();
      } catch (IOException e) {
        ExceptionHandler.handleExceptionExit(e);
      }
    }
    return currentValue;
  }

  /**
   * Creates a new trial
   * 
   * @param randomSeed
   *          the <code>long</code> value representing the random seed to use for the new trial
   * 
   * @return the <code>int<code> trial ID; returns 0 if trial does not exist in database
   */
  public synchronized int newTrial(long randomSeed) {
    PreparedStatement createNewTrial = null;
    try {
      createNewTrial = this.db.prepareStatement("insert into trial(random_seed) values (?)");
      createNewTrial.setLong(1, randomSeed);
      createNewTrial.executeUpdate();
      createNewTrial.close();
    } catch (SQLException e) {
      ExceptionHandler.handleSQLException(e, this.db, createNewTrial);
    }

    PreparedStatement lastTrialID = null;
    int currentValue = 0;
    try {
      lastTrialID = this.db.prepareStatement("select currval('trial_trial_id_seq')");
      this.results = lastTrialID.executeQuery();
      if (this.results.next()) {
        currentValue = this.results.getInt("currval");
      }
      lastTrialID.close();
    } catch (SQLException e) {
      ExceptionHandler.handleSQLException(e, this.db, lastTrialID);
    }
    return currentValue;
  }

  /**
   * Provides the number of file matches
   * 
   * @return the <code>int</code> count of file matches
   */
  public int getMatchingFileCount() {
    PreparedStatement countMatchingFiles = null;
    int currentValue = 0;
    try {
      countMatchingFiles = this.db.prepareStatement("select count(*) from file_match");
      this.results = countMatchingFiles.executeQuery();
      if (this.results.next()) {
        currentValue = this.results.getInt(1);
      }
      countMatchingFiles.close();
    } catch (SQLException e) {
      ExceptionHandler.handleSQLException(e, this.db, countMatchingFiles);
    }
    return currentValue;
  }

  /**
   * Provides the ID of the tokenizer or 0 if it is not in the tokenizer table of this database
   * 
   * @param tokenizer
   *          the string name of the tokenizer
   * 
   * @return the <code>int</code> tokenizer ID; returns 0 if tokenizer does not exist in database
   */
  public int getTokenizerID(String tokenizer) {
    PreparedStatement findTokenizerID = null;
    int currentValue = 0;
    try {
      findTokenizerID = this.db.prepareStatement("select tokenizer_id from tokenizer where tokenizer_name=(?)");
      findTokenizerID.setString(1, tokenizer);
      this.results = findTokenizerID.executeQuery();

      if (this.results.next()) {
        currentValue = this.results.getInt("tokenizer_id");
      }
      findTokenizerID.close();
    } catch (SQLException e) {
      ExceptionHandler.handleSQLException(e, this.db, findTokenizerID);
    }
    return currentValue;
  }

  /**
   * Adds a tokenizer to the tokenizer table in this database
   * 
   * @param tokenizer
   *          the string name of the tokenizer
   * 
   * @return the <code>int</code> tokenizer ID; returns 0 if tokenizer does not exist in database
   */
  public synchronized int addTokenizer(String tokenizer) {
    int tokenizerID = getTokenizerID(tokenizer);
    if (0 == tokenizerID) {
      PreparedStatement insertTokenizer = null;
      try {
        insertTokenizer = this.db.prepareStatement("insert into tokenizer (tokenizer_name) values (?)");
        insertTokenizer.setString(1, tokenizer);
        insertTokenizer.executeUpdate();

        this.results = this.sql.executeQuery("select currval('tokenizer_tokenizer_id_seq')");
        if (this.results.next()) {
          tokenizerID = this.results.getInt("currval");
        }
        insertTokenizer.close();
      } catch (SQLException e) {
        ExceptionHandler.handleSQLException(e, this.db, insertTokenizer);
      }
    }
    return tokenizerID;
  }

  /**
   * Adds a list of tokenizers to an experiment
   * 
   * @param experimentID
   *          the <code>int</code> experiment ID to look up in the database
   * @param tokenizerVec
   *          the list of tokenizer strings
   */
  public void saveTokenizers(int experimentID, List<String> tokenizerVec) {
    for (String tokenizer : tokenizerVec) {
      int tokenizerID = getTokenizerID(tokenizer);
      if (0 == tokenizerID) {
        tokenizerID = addTokenizer(tokenizer);
      }

      // add the mapping to the trial_tokenizer table
      PreparedStatement addTokenizerToExperiment = null;
      try {
        addTokenizerToExperiment = this.db
            .prepareStatement("insert into experiment_tokenizer (experiment_id, tokenizer_id) values(?,?)");
        addTokenizerToExperiment.setInt(1, experimentID);
        addTokenizerToExperiment.setInt(2, tokenizerID);
        addTokenizerToExperiment.executeUpdate();
        addTokenizerToExperiment.close();
      } catch (SQLException e) {
        ExceptionHandler.handleSQLException(e, this.db, addTokenizerToExperiment);
      }
    }
  }

  /**
   * Drops the indexes on the fingerprint_file_id, fingerprint_trial_id, and fingerprint_mangler_id tables
   */
  public void startBulkFingerprintInsertion() {
    try {
      this.sql.executeUpdate("drop index fingerprint_file_id");
      this.sql.executeUpdate("drop index fingerprint_trial_id");
      this.sql.executeUpdate("drop index fingerprint_mangler_id");
    } catch (SQLException e) {
      ExceptionHandler.handleSQLException(e, this.db, this.sql);
    }
  }

  /**
   * Rebuild indexes for fingerprint_file_id, fingerprint_trial_id, and fingerprint_mangler_id tables
   */
  public void finishBulkFingerprintInsertion() {
    try {
      this.sql.executeUpdate("create index fingerprint_file_id on fingerprint using hash (file_id);");
      this.sql.executeUpdate("create index fingerprint_trial_id on fingerprint using hash (trial_id);");
      this.sql.executeUpdate("create index fingerprint_mangler_id on fingerprint using hash (mangler_id);");
    } catch (SQLException e) {
      ExceptionHandler.handleSQLException(e, this.db, this.sql);
    }
  }

  /**
   * Inserts fingerprint information into the database and returns the corresponding fingerprint ID
   * 
   * @param file
   *          the file that was fingerprinted
   * @param fingerprint
   *          the <code>byte</code> array containing the fingerprint to insert
   * @param trialID
   *          the <code>int</code> trial ID to insert
   * @param manglerID
   *          the <code>int</code> mangler ID to insert
   * 
   * @return the <code>int</code> fingerprint ID; returns 0 if fingerprint does not exist in database
   */
  public synchronized int insertFingerprint(File file, byte[] fingerprint, int trialID, int manglerID) {
    int fingerprintID = 0;
    int fileID = addFile(file);

    PreparedStatement addFingerprint = null;
    try {
      addFingerprint = this.db
          .prepareStatement("insert into fingerprint(file_id, trial_id, mangler_id, fingerprint) values (?,?,?,?)");
      addFingerprint.setInt(1, fileID);
      addFingerprint.setInt(2, trialID);
      addFingerprint.setInt(3, manglerID);
      addFingerprint.setBytes(4, fingerprint);
      addFingerprint.executeUpdate();
      addFingerprint.close();
    } catch (SQLException e) {
      ExceptionHandler.handleSQLException(e, this.db, addFingerprint);
    }

    PreparedStatement lastFingerprintID = null;
    try {
      lastFingerprintID = this.db.prepareStatement("select currval('fingerprint_fingerprint_id_seq')");
      this.results = lastFingerprintID.executeQuery();

      if (this.results.next()) {
        fingerprintID = this.results.getInt("currval");
      }
      lastFingerprintID.close();
    } catch (SQLException e) {
      ExceptionHandler.handleSQLException(e, this.db, lastFingerprintID);
    }
    return fingerprintID;
  }

  /**
   * Provides the ID for the existing mangler setting that matches or creates a new one
   * 
   * @param manglerSettings
   *          the string mangler settings
   * 
   * @return the <code>int</code> mangler ID; returns 0 if mangler does not exist in database
   */
  public synchronized int getManglerID(String manglerSettings) {

    if (manglerSettings == null || manglerSettings.equalsIgnoreCase("null")) {
      return 1;
    }

    int manglerID = 0;
    double delete_token = 0.0;
    double delete_char = 0.0;
    double delete_whitespace = 0.0;
    double change_to_char = 0.0;
    double change_token = 0.0;
    double split_token = 0.0;
    double change_to_punc = 0.0;
    double delete_section = 0.0;
    double delete_head = 0.0;
    double delete_tail = 0.0;
    double summary_reduce = 0.0;

    PreparedStatement getManglerID = null;
    try {
      StringTokenizer st = new StringTokenizer(manglerSettings);
      List<String> tokens = new Vector<String>();
      while (st.hasMoreTokens()) {
        tokens.add(st.nextToken());
      }

      if (manglerSettings.contains("deletetoken")) {
        delete_token = Double.valueOf(tokens.get(1).trim());
      }

      if (manglerSettings.contains("deletechar")) {
        delete_char = Double.valueOf(tokens.get(1).trim());
      }

      if (manglerSettings.contains("deletewhitespace")) {
        delete_whitespace = Double.valueOf(tokens.get(1).trim());
      }

      if (manglerSettings.contains("changetochar")) {
        change_to_char = Double.valueOf(tokens.get(1).trim());
      }

      if (manglerSettings.contains("changetoken")) {
        change_token = Double.valueOf(tokens.get(1).trim());
      }

      if (manglerSettings.contains("splittoken")) {
        split_token = Double.valueOf(tokens.get(1).trim());
      }

      if (manglerSettings.contains("changetopunc")) {
        change_to_punc = Double.valueOf(tokens.get(1).trim());
      }

      if (manglerSettings.contains("deletesection")) {
        delete_section = Double.valueOf(tokens.get(1).trim());
      }

      if (manglerSettings.contains("deletehead")) {
        delete_head = Double.valueOf(tokens.get(1).trim());
      }

      if (manglerSettings.contains("deletetail")) {
        delete_tail = Double.valueOf(tokens.get(1).trim());
      }

      if (manglerSettings.contains("summaryreduce")) {
        summary_reduce = Double.valueOf(tokens.get(1).trim());
      }

      getManglerID = this.db
          .prepareStatement("select mangler_id from mangler where delete_token = (?) and  "
              + "delete_char= (?) and delete_whitespace = (?) and change_to_char = (?) and change_token = (?) and split_token = (?) "
              + "and change_to_punc = (?) and delete_section = (?) and delete_head = (?) and delete_tail = (?) and summary_reduce = (?)");
      getManglerID.setDouble(1, delete_token);
      getManglerID.setDouble(2, delete_char);
      getManglerID.setDouble(3, delete_whitespace);
      getManglerID.setDouble(4, change_to_char);
      getManglerID.setDouble(5, change_token);
      getManglerID.setDouble(6, split_token);
      getManglerID.setDouble(7, change_to_punc);
      getManglerID.setDouble(8, delete_section);
      getManglerID.setDouble(9, delete_head);
      getManglerID.setDouble(10, delete_tail);
      getManglerID.setDouble(11, summary_reduce);
      this.results = getManglerID.executeQuery();

      if (this.results.next()) {
        manglerID = this.results.getInt("mangler_id");
      }
      getManglerID.close();
    } catch (SQLException e) {
      ExceptionHandler.handleSQLException(e, this.db, getManglerID);
    }

    PreparedStatement newManglerID = null;
    try {
      if (manglerID == 0) {
        // Create a new entry
        String manglerSetting[] = manglerSettings.trim().split("\\s+");
        String manglerName = manglerSetting[0];
        double manglerValue = Double.parseDouble(manglerSetting[1]);

        newManglerID = this.db.prepareStatement("insert into mangler (mangler_name,mangler_setting,"
            + "delete_token,delete_char, delete_whitespace, change_to_char, change_token, split_token, change_to_punc,"
            + "delete_section, delete_head, delete_tail, summary_reduce) values (?,?,?,?,?,?,?,?,?,?,?,?,?)");
        newManglerID.setString(1, manglerName);
        newManglerID.setDouble(2, manglerValue);
        newManglerID.setDouble(3, delete_token);
        newManglerID.setDouble(4, delete_char);
        newManglerID.setDouble(5, delete_whitespace);
        newManglerID.setDouble(6, change_to_char);
        newManglerID.setDouble(7, change_token);
        newManglerID.setDouble(8, split_token);
        newManglerID.setDouble(9, change_to_punc);
        newManglerID.setDouble(10, delete_section);
        newManglerID.setDouble(11, delete_head);
        newManglerID.setDouble(12, delete_tail);
        newManglerID.setDouble(13, summary_reduce);
        newManglerID.executeUpdate();
        newManglerID.close();

        PreparedStatement lastManglerID = null;
        try {
          lastManglerID = this.db.prepareStatement("select currval('mangler_mangler_id_seq')");
          this.results = lastManglerID.executeQuery();

          if (this.results.next()) {
            manglerID = this.results.getInt("currval");
          }
          lastManglerID.close();
        } catch (SQLException e) {
          ExceptionHandler.handleSQLException(e, this.db, lastManglerID);
        }
      }
    } catch (SQLException e) {
      ExceptionHandler.handleSQLException(e, this.db, newManglerID);
    }
    return manglerID;
  }

  /**
   * Adds the mangler ID to the trial in this database
   * 
   * @param trialID
   *          the <code>int</code> trial ID to insert
   * @param manglerID
   *          the <code>int</code> mangler ID to insert
   */
  public void addManglerToTrial(int trialID, int manglerID) {
    PreparedStatement insertManglerTrialEntry = null;
    try {
      insertManglerTrialEntry = this.db
          .prepareStatement("insert into trial_mangler (trial_id, mangler_id) values (?,?)");
      insertManglerTrialEntry.setInt(1, trialID);
      insertManglerTrialEntry.setInt(2, manglerID);
      insertManglerTrialEntry.executeUpdate();
      insertManglerTrialEntry.close();
    } catch (SQLException e) {
      ExceptionHandler.handleSQLException(e, this.db, insertManglerTrialEntry);
    }
  }

  /**
   * Adds the trial ID to the experiment in this database
   * 
   * @param experimentID
   *          the <code>int</code> experiment ID to look up in the database
   * @param trialID
   *          the <code>int</code> trial ID to insert
   */
  public void addTrialToExperiment(int experimentID, int trialID) {
    PreparedStatement insertTrialInExperiment = null;
    try {
      insertTrialInExperiment = this.db
          .prepareStatement("insert into experiment_trial (experiment_id, trial_id) values (?,?)");
      insertTrialInExperiment.setInt(1, experimentID);
      insertTrialInExperiment.setInt(2, trialID);
      insertTrialInExperiment.executeUpdate();
      insertTrialInExperiment.close();
    } catch (SQLException e) {
      ExceptionHandler.handleSQLException(e, this.db, insertTrialInExperiment);
    }
  }

  /**
   * Adds a "null" mangler to position 1 of the trial_mangler table
   * 
   * @param trialID
   *          the <code>int</code> trial ID to insert
   */
  public void addNullManglerToTrial(int trialID) {
    int nullPos = 1;
    PreparedStatement insertManglerTrialEntry = null;
    try {
      insertManglerTrialEntry = this.db
          .prepareStatement("insert into trial_mangler (trial_id, mangler_id) values (?,?)");
      insertManglerTrialEntry.setInt(1, trialID);
      insertManglerTrialEntry.setInt(2, nullPos);
      insertManglerTrialEntry.executeUpdate();
      insertManglerTrialEntry.close();
    } catch (SQLException e) {
      ExceptionHandler.handleSQLException(e, this.db, insertManglerTrialEntry);
    }
  }

  /**
   * Provides all the mangler IDs for a trial
   * 
   * @param trialID
   *          the <code>int</code> trial ID to load
   * 
   * @return the list of integer mangler IDs
   */
  public List<Integer> getManglersForTrial(int trialID) {
    List<Integer> trialManglers = new Vector<Integer>();
    PreparedStatement findManglerFromTrial = null;
    try {
      findManglerFromTrial = this.db
          .prepareStatement("select mangler_id from trial_mangler where trial_id=(?) order by trial_mangler_id");
      findManglerFromTrial.setInt(1, trialID);
      this.results = findManglerFromTrial.executeQuery();
      while (this.results.next()) {
        trialManglers.add(this.results.getInt("mangler_id"));
      }
      this.results = null;
      findManglerFromTrial.close();
    } catch (SQLException e) {
      ExceptionHandler.handleSQLException(e, this.db, findManglerFromTrial);
    }
    return trialManglers;
  }

  /**
   * Provides all the file IDs and whether they have matches in a trial
   * 
   * @param trialID
   *          the <code>int</code> trial ID to load
   * 
   * @return the map of integer file IDs and their boolean values indicating if the have a match
   */
  public Map<Integer, Boolean> getTrialFileIDs(int trialID) {
    Map<Integer, Boolean> trialFileIDs = new HashMap<Integer, Boolean>();
    PreparedStatement findTrialFileIDs = null;
    try {
      findTrialFileIDs = this.db.prepareStatement("select distinct(file_id) from fingerprint where trial_id = ?");
      findTrialFileIDs.setInt(1, trialID);
      this.results = findTrialFileIDs.executeQuery();
      while (this.results.next()) {
        trialFileIDs.put(this.results.getInt("file_id"), true);
      }
      findTrialFileIDs.close();
    } catch (SQLException e) {
      ExceptionHandler.handleSQLException(e, this.db, findTrialFileIDs);
    }
    return trialFileIDs;
  }

  /**
   * Provides all the file IDs of files that have matches
   * 
   * @param fileID
   *          the <code>int</code> file ID to load
   * 
   * @return the integer hashSet containing all the matching file IDs
   */
  public Set<Integer> getAllMatchingFileIDs(int fileID) {
    Set<Integer> fileIDs = new HashSet<Integer>();
    PreparedStatement findAllFileMatchIDs = null;
    try {
      findAllFileMatchIDs = this.db.prepareStatement("select match_id from file_match where file_id = ?");
      findAllFileMatchIDs.setInt(1, fileID);
      this.results = findAllFileMatchIDs.executeQuery();
      while (this.results.next()) {
        fileIDs.add(this.results.getInt("match_id"));
      }
      this.results = null;
      findAllFileMatchIDs.close();
    } catch (SQLException e) {
      ExceptionHandler.handleSQLException(e, this.db, findAllFileMatchIDs);
    }
    return fileIDs;
  }

  /**
   * Provides the fingerprints and fingerprint IDs with null manglers for a trial
   * 
   * @param trialID
   *          the <code>int</code> trial ID to load
   * 
   * @return the hashMap of integer fingerprint IDs and their byte array fingerprints
   */
  public Map<Integer, byte[]> getOriginalFingerprintsFromTrial(int trialID) {
    Map<Integer, byte[]> trialFingerprints = new HashMap<Integer, byte[]>();
    PreparedStatement findOriginalFingerprintsFromTrial = null;
    try {
      findOriginalFingerprintsFromTrial = this.db
          .prepareStatement("select fingerprint_id, fingerprint from fingerprint where trial_id=(?) and mangler_id='1'");
      findOriginalFingerprintsFromTrial.setInt(1, trialID);
      this.results = findOriginalFingerprintsFromTrial.executeQuery();

      while (this.results.next()) {
        trialFingerprints.put(this.results.getInt("fingerprint_id"), this.results.getBytes("fingerprint"));
      }
      this.results = null;
      findOriginalFingerprintsFromTrial.close();
    } catch (SQLException e) {
      ExceptionHandler.handleSQLException(e, this.db, findOriginalFingerprintsFromTrial);
    }
    return trialFingerprints;
  }

  /**
   * Provides all the fingerprints and fingerprint IDs for a trial
   * 
   * @param trialID
   *          the <code>int</code> trial ID to load
   * 
   * @return the hashMap of integer fingerprint IDs and their byte array fingerprints
   */
  public Map<Integer, byte[]> getAllFingerprintsFromTrial(int trialID) {
    Map<Integer, byte[]> trialFingerprints = new HashMap<Integer, byte[]>();
    PreparedStatement findAllFingerprintsFromTrial = null;
    try {
      findAllFingerprintsFromTrial = this.db
          .prepareStatement("select fingerprint_id, fingerprint from fingerprint where trial_id=(?)");
      findAllFingerprintsFromTrial.setInt(1, trialID);
      this.results = findAllFingerprintsFromTrial.executeQuery();

      while (this.results.next()) {
        trialFingerprints.put(this.results.getInt("fingerprint_id"), this.results.getBytes("fingerprint"));
      }
      this.results = null;
      findAllFingerprintsFromTrial.close();
    } catch (SQLException e) {
      ExceptionHandler.handleSQLException(e, this.db, findAllFingerprintsFromTrial);
    }
    return trialFingerprints;
  }

  /**
   * Adds the two matching fingerprints, if not already stored, to fingerprintMatchCache
   * 
   * @param sourceFingerprintID
   *          the <code>int</code> fingerprint id to insert into the fingerprint_match table
   * @param targetFingerprintID
   *          the <code>int</code> fingerprint id to insert into the fingerprint_match table
   */
  private void addToFingerprintMatchCache(int sourceFingerprintID, int targetFingerprintID) {
    if (this.fingerprintMatchCache.containsKey(sourceFingerprintID)) {
      HashMap<Integer, Integer> existingHash = this.fingerprintMatchCache.get(sourceFingerprintID);
      if (!existingHash.containsKey(targetFingerprintID)) {
        existingHash.put(targetFingerprintID, null);
      }
    } else { // else the match is already stored in the cache
      HashMap<Integer, Integer> newHash = new HashMap<Integer, Integer>();
      newHash.put(targetFingerprintID, null);
      this.fingerprintMatchCache.put(sourceFingerprintID, newHash);
    }
  }

  /**
   * Checks whether the two fingerprints exist as matches in the fingerprintMatchCache
   * 
   * @param sourceFingerprintID
   *          the <code>int</code> fingerprint id to insert into the fingerprint_match table
   * @param targetFingerprintID
   *          the <code>int</code> fingerprint id to insert into the fingerprint_match table
   * 
   * @return the <code>boolean</code> whether the fingerprints exist as matches
   */
  private boolean isFingerprintMatchInCache(int sourceFingerprintID, int targetFingerprintID) {
    if (this.fingerprintMatchCache == null || !this.fingerprintMatchCache.containsKey(sourceFingerprintID)) {
      return false;
    } else if (this.fingerprintMatchCache.get(sourceFingerprintID).containsKey(targetFingerprintID)) {
      return true;
    }
    return false;
  }

  /**
   * Loads existing matches from the database into the fingerprint match cache
   * 
   * @param trialID
   *          the <code>int</code> trial ID to load
   */
  private void loadFingerprintMatchCache(int trialID) {
    int sourceID, matchID;
    PreparedStatement getAllFingerprintMatches = null;
    try {
      getAllFingerprintMatches = this.db
          .prepareStatement("select source_print_id, target_print_id from fingerprint_match where trial_id = ?");
      getAllFingerprintMatches.setInt(1, trialID);
      this.results = getAllFingerprintMatches.executeQuery();

      while (this.results.next()) {
        sourceID = this.results.getInt(1);
        matchID = this.results.getInt(2);
        addToFingerprintMatchCache(sourceID, matchID);
      }
      this.results = null;
      getAllFingerprintMatches.close();
    } catch (SQLException e) {
      ExceptionHandler.handleSQLException(e, this.db, getAllFingerprintMatches);
    }
  }

  /**
   * Drops the indexes on the fingerprint_match tables
   * 
   * @param trialID
   *          the <code>int</code> trial ID to drop
   */
  public void startBulkFingerprintMatchInsertion(int trialID) {

    loadFingerprintMatchCache(trialID);
    try {
      this.sql.executeUpdate("drop index fingerprint_match_trial_id");
      this.sql.executeUpdate("drop index fingerprint_match_source_print_id");
      this.sql.executeUpdate("drop index fingerprint_match_target_print_id");
      this.sql.executeUpdate("drop index fingerprint_match_source_mangler_id");
    } catch (SQLException e) {
      // An SQL exception here just means that the index isn't built, and
      // therefore
      // can't be dropped, so ignore it.
    }
  }

  /**
   * Adds fingerprint match information if the match is not already recorded in the fingerprint_match table of this
   * database
   * 
   * @param trialID
   *          the <code>int</code> trial ID to insert into the fingerprint_match table
   * @param sourceFingerprintID
   *          the <code>int</code> fingerprint id to insert into the fingerprint_match table
   * @param targetFingerprintID
   *          the <code>int</code> fingerprint id to insert into the fingerprint_match table
   * @param doesMatch
   *          the <code>boolean</code> whether the experiment matcher found the two fingerprints to have matching files
   * @param shouldMatch
   *          the <code>boolean</code> whether the two fingerprints have matching files
   * @param score
   *          the similarity score for the two fingerprints
   */
  public void addToFingerprintMatch(int trialID, int sourceFingerprintID, int targetFingerprintID, boolean doesMatch,
      boolean shouldMatch, int score) {

    if (!isFingerprintMatchInCache(sourceFingerprintID, targetFingerprintID)) {
      try {
        this.insertFingerprintMatch.setInt(1, trialID);
        this.insertFingerprintMatch.setInt(2, sourceFingerprintID);
        this.insertFingerprintMatch.setInt(3, targetFingerprintID);
        this.insertFingerprintMatch.setBoolean(4, doesMatch);
        this.insertFingerprintMatch.setBoolean(5, shouldMatch);
        this.insertFingerprintMatch.setInt(6, sourceFingerprintID);
        this.insertFingerprintMatch.setInt(7, targetFingerprintID);
        this.insertFingerprintMatch.setInt(8, score);
        this.insertFingerprintMatch.addBatch();
        addToFingerprintMatchCache(sourceFingerprintID, targetFingerprintID);
        this.fingerprintMatchAddedCount++;

        // Fingerprints match both ways, so reverse the fingerprint order if
        // both IDS are not the same
        if (targetFingerprintID != sourceFingerprintID) {
          this.insertFingerprintMatch.setInt(1, trialID);
          this.insertFingerprintMatch.setInt(2, targetFingerprintID);
          this.insertFingerprintMatch.setInt(3, sourceFingerprintID);
          this.insertFingerprintMatch.setBoolean(4, doesMatch);
          this.insertFingerprintMatch.setBoolean(5, shouldMatch);
          this.insertFingerprintMatch.setInt(6, targetFingerprintID);
          this.insertFingerprintMatch.setInt(7, sourceFingerprintID);
          this.insertFingerprintMatch.setInt(8, score);
          this.insertFingerprintMatch.addBatch();
          addToFingerprintMatchCache(targetFingerprintID, sourceFingerprintID);
          this.fingerprintMatchAddedCount++;
        }

        if ((this.fingerprintMatchAddedCount % this.batchSize) == 0) {
          Global.monitor.printMessage("DBInterface", "Inserting Fingerprint Matches, up to: "
              + this.fingerprintMatchAddedCount);
          this.insertFingerprintMatch.executeBatch();
        }
      } catch (SQLException e) {
        ExceptionHandler.handleSQLException(e);
      }
    }
  }

  /**
   * Rebuilds indexes on the fingerprint_match tables
   */
  public void finishBulkFingerprintMatchInsertion() {
    try {
      // get any remaining records
      if (this.fingerprintMatchAddedCount > 0) {
        this.insertFingerprintMatch.executeBatch();
      }
    } catch (SQLException e) {
      ExceptionHandler.handleSQLException(e, this.db, this.insertFingerprintMatch);
    }

    try {
      // Rebuild the indexes
      this.sql.executeUpdate("create index fingerprint_match_trial_id on fingerprint using hash (trial_id)");
      this.sql
          .executeUpdate("create index fingerprint_match_source_print_id on fingerprint_match using hash (source_print_id)");
      this.sql
          .executeUpdate("create index fingerprint_match_target_print_id on fingerprint_match using hash (target_print_id)");
      this.sql
          .executeUpdate("create index fingerprint_match_source_mangler_id on fingerprint_match using hash (source_mangler_id)");
    } catch (SQLException e) {
      ExceptionHandler.handleSQLException(e, this.db, this.sql);
    }
  }

  /**
   * Sets all the matches for each fingerprint ID in this database
   * 
   * @param fingerprintID
   *          the <code>int</code> fingerprint ID to look up in the database
   */
  public void setTrueMatches(int fingerprintID) {
    PreparedStatement updateTrueMatches = null;
    try {
      updateTrueMatches = this.db
          .prepareStatement("update fingerprint_match set should_match = 't' where "
              + "source_print_id=(?) and target_print_id in (select fingerprint_id from fingerprint where file_id in "
              + "(select match_id from file_match where file_id in (select file_id from fingerprint where fingerprint_id = (?))))");
      updateTrueMatches.setInt(1, fingerprintID);
      updateTrueMatches.setInt(2, fingerprintID);
      updateTrueMatches.executeUpdate();
      updateTrueMatches.close();
    } catch (SQLException e) {
      ExceptionHandler.handleSQLException(e, this.db, updateTrueMatches);
    }
  }

  /**
   * Counts the number of entries with the given manglers and matching results in the fingerprint_match table for a
   * trial
   * 
   * @param trialID
   *          the <code>int</code> trial ID to load from the fingerprint_match table
   * @param sourceMangler
   *          the <code>int</code> mangler id to load from the fingerprint_match table
   * @param targetMangler
   *          the <code>int</code> mangler id to load from the fingerprint_match table
   * @param shouldMatch
   *          the <code>boolean</code> whether the two fingerprints have matching files (to load from the
   *          fingerprint_match table)
   * @param doesMatch
   *          the <code>boolean</code> whether the experiment matcher found the two fingerprints to have matching files
   *          (to load from the fingerprint_match table)
   * 
   * @return the <code>int</code> number of results found for the given manglers and matching results
   */
  public int countResults(int trialID, int sourceMangler, int targetMangler, boolean shouldMatch, boolean doesMatch) {
    PreparedStatement countTrialResults = null;
    int count = 0;
    try {
      countTrialResults = this.db.prepareStatement("select count(*) from fingerprint_match where trial_id = (?) "
          + "and source_mangler_id = (?) and target_mangler_id=(?) and should_match=(?) and does_match=(?)");
      countTrialResults.setInt(1, trialID);
      countTrialResults.setInt(2, sourceMangler);
      countTrialResults.setInt(3, targetMangler);
      countTrialResults.setBoolean(4, shouldMatch);
      countTrialResults.setBoolean(5, doesMatch);
      this.results = countTrialResults.executeQuery();

      if (this.results.next()) {
        count = this.results.getInt(1);
      }
      countTrialResults.close();
    } catch (SQLException e) {
      ExceptionHandler.handleSQLException(e, this.db, countTrialResults);
    }
    return count;
  }

  /**
   * Adds the results from a trial to this database
   * 
   * @param trialID
   *          the <code>int</code> trial ID to insert
   * @param manglerID
   *          the <code>int</code> mangler ID to insert
   * @param truePositive
   *          the <code>int</code> count of true positives for the mangler in this trial
   * @param trueNegative
   *          the <code>int</code> count of true negatives for the mangler in this trial
   * @param falsePositive
   *          the <code>int</code> count of false positives for the mangler in this trial
   * @param falseNegative
   *          the <code>int</code> count of true negatives for the mangler in this trial
   * @param precision
   *          the <code>double</code> precision for the mangler in this trial
   * @param recall
   *          the <code>double</code> recall for the mangler in this trial
   * @param fscore
   *          the <code>double</code> fscore for the mangler in this trial
   */
  public void addTrialResults(int trialID, int manglerID, int truePositive, int trueNegative, int falsePositive,
      int falseNegative, double precision, double recall, double fscore) {
    PreparedStatement insertTrialResults = null;
    try {
      insertTrialResults = this.db
          .prepareStatement("insert into trial_results (trial_id, print_mangler_id, true_positive_count, "
              + "true_negative_count, false_positive_count, false_negative_count, precision, recall, fscore) values (?,?,?,?,?,?,?,?,?)");
      insertTrialResults.setInt(1, trialID);
      insertTrialResults.setInt(2, manglerID);
      insertTrialResults.setInt(3, truePositive);
      insertTrialResults.setInt(4, trueNegative);
      insertTrialResults.setInt(5, falsePositive);
      insertTrialResults.setInt(6, falseNegative);
      insertTrialResults.setDouble(7, precision);
      insertTrialResults.setDouble(8, recall);
      insertTrialResults.setDouble(9, fscore);
      insertTrialResults.executeUpdate();
      insertTrialResults.close();
    } catch (SQLException e) {
      ExceptionHandler.handleSQLException(e, this.db, insertTrialResults);
    }
  }

  /**
   * Extracts a dictionary from a trial in this database
   * 
   * @param trialID
   *          the <code>int</code> trial ID to load
   * @param dict_file
   *          the string filename to write the dictionary to
   */
  public void extractDictionary(int trialID, String dict_file) {
    PreparedStatement dictionaryOIDFromTrial = null;
    try {
      this.db.setAutoCommit(false);
      dictionaryOIDFromTrial = this.db
          .prepareStatement("select stored_dictionary from dictionary where dictionary_id = (select dictionary_id from trial where trial_id=(?))");
      dictionaryOIDFromTrial.setInt(1, trialID);
      this.results = dictionaryOIDFromTrial.executeQuery();
      if (this.results.next()) {
        Blob b = this.results.getBlob(1);
        InputStream input = b.getBinaryStream();
        BufferedOutputStream os = new BufferedOutputStream(new FileOutputStream(dict_file));
        int c;
        while ((c = input.read()) != -1) {
          os.write(c);
        }
        os.flush();
        os.close();
        input.close();
      }
      this.db.setAutoCommit(true);
      dictionaryOIDFromTrial.close();
    } catch (SQLException e) {
      ExceptionHandler.handleSQLException(e, this.db, dictionaryOIDFromTrial);
    } catch (FileNotFoundException e) {
      ExceptionHandler.handleExceptionExit(e);
    } catch (IOException e) {
      ExceptionHandler.handleExceptionExit(e);
    }
  }

  /**
   * Extracts a configuration file from an experiment in this database
   * 
   * @param experimentID
   *          the <code>int</code> experiment ID to load
   * @param config_file
   *          the string filename to write the configuration file to
   */
  public void extractConfigFile(int experimentID, String config_file) {
    PreparedStatement configOIDFromExperiment = null;
    try {
      this.db.setAutoCommit(false);
      configOIDFromExperiment = this.db
          .prepareStatement("select config_file from experiment where experiment_id = (?)");
      configOIDFromExperiment.setInt(1, experimentID);
      this.results = configOIDFromExperiment.executeQuery();
      if (this.results.next()) {
        Blob b = this.results.getBlob(1);
        InputStream input = b.getBinaryStream();
        BufferedOutputStream os = new BufferedOutputStream(new FileOutputStream(config_file));
        int c;
        while ((c = input.read()) != -1) {
          os.write(c);
        }
        os.flush();
        os.close();
        input.close();
      }
      this.db.setAutoCommit(true);
      configOIDFromExperiment.close();
    } catch (SQLException e) {
      ExceptionHandler.handleSQLException(e, this.db, configOIDFromExperiment);
    } catch (FileNotFoundException e) {
      ExceptionHandler.handleExceptionExit(e);
    } catch (IOException e) {
      ExceptionHandler.handleExceptionExit(e);
    }
  }

  /**
   * Provides the description for an experiment
   * 
   * @param experimentID
   *          the <code>int</code> experiment ID to look up in the database
   * 
   * @return the string description; returns an empty string if a description is not present in the database for the
   *         provided experiment
   */
  public String getExperimentDescription(int experimentID) {
    PreparedStatement getExpDescription = null;
    String desc = "";
    try {
      getExpDescription = this.db.prepareStatement("select description from experiment where experiment_id = (?)");
      getExpDescription.setInt(1, experimentID);
      this.results = getExpDescription.executeQuery();
      if (this.results.next()) {
        desc = this.results.getString(1);
      }
      getExpDescription.close();
    } catch (SQLException e) {
      ExceptionHandler.handleSQLException(e, this.db, getExpDescription);
    }
    return desc;
  }

  /**
   * Provides the path of the dataset used in an experiment
   * 
   * @param experimentID
   *          the <code>int</code> experiment ID to look up in the database
   * 
   * @return the string dataset path; returns an empty string if a dataset path is not present in the database for the
   *         provided experiment
   */
  public String getExperimentDatasetPath(int experimentID) {
    PreparedStatement getExpDatasetPath = null;
    String path = "";
    try {
      getExpDatasetPath = this.db.prepareStatement("select dataset_path from experiment where experiment_id = (?)");
      getExpDatasetPath.setInt(1, experimentID);
      this.results = getExpDatasetPath.executeQuery();
      if (this.results.next()) {
        path = this.results.getString(1);
      }
      getExpDatasetPath.close();
    } catch (SQLException e) {
      ExceptionHandler.handleSQLException(e, this.db, getExpDatasetPath);
    }
    return path;
  }

  /**
   * Provides the fingerprinter name used in an experiment
   * 
   * @param experimentID
   *          the <code>int</code> experiment ID to look up in the database
   * 
   * @return the string fingerprinter name; returns an empty string if a fingerprinter name is not present in the
   *         database for the provided experiment
   */
  public String getExperimentFingerprinterName(int experimentID) {
    PreparedStatement getExpFingerprinterName = null;
    String fingerprint = "";
    try {
      getExpFingerprinterName = this.db
          .prepareStatement("select fingerprinter_name from experiment where experiment_id = (?)");
      getExpFingerprinterName.setInt(1, experimentID);
      this.results = getExpFingerprinterName.executeQuery();
      if (this.results.next()) {
        fingerprint = this.results.getString(1);
      }
      getExpFingerprinterName.close();
    } catch (SQLException e) {
      ExceptionHandler.handleSQLException(e, this.db, getExpFingerprinterName);
    }
    return fingerprint;
  }

  /**
   * Provides the matcher name used in an experiment
   * 
   * @param experimentID
   *          the <code>int</code> experiment ID to look up in the database
   * 
   * @return the string matcher name; returns an empty string if a matcher name is not present in the database for the
   *         provided experiment
   */
  public String getExperimentMatcherName(int experimentID) {
    PreparedStatement getExpMatcherName = null;
    String matcher = "";
    try {
      getExpMatcherName = this.db.prepareStatement("select matcher_name from experiment where experiment_id = (?)");
      getExpMatcherName.setInt(1, experimentID);
      this.results = getExpMatcherName.executeQuery();
      if (this.results.next()) {
        matcher = this.results.getString(1);
      }
      getExpMatcherName.close();
    } catch (SQLException e) {
      ExceptionHandler.handleSQLException(e, this.db, getExpMatcherName);
    }
    return matcher;
  }

  /**
   * Provides the minimum score used for the matcher in an experiment
   * 
   * @param experimentID
   *          the <code>int</code> experiment ID to look up in the database
   * 
   * @return the <code>int</code> matcher minimum score parameter
   */
  public int getExperimentMatcherParameter(int experimentID) {
    PreparedStatement getExpMatcherParameter = null;
    int parameter = 0;
    try {
      getExpMatcherParameter = this.db
          .prepareStatement("select matcher_parameter from experiment where experiment_id = (?)");
      getExpMatcherParameter.setInt(1, experimentID);
      this.results = getExpMatcherParameter.executeQuery();
      if (this.results.next()) {
        parameter = this.results.getInt(1);
      }
      getExpMatcherParameter.close();
    } catch (SQLException e) {
      ExceptionHandler.handleSQLException(e, this.db, getExpMatcherParameter);
    }
    return parameter;
  }

  /**
   * Provides the minimum normalized IDF for an experiment's dictionary
   * 
   * @param experimentID
   *          the <code>int</code> experiment ID to look up in the database
   * 
   * @return the <code>double</code> minimum normalized IDF used in the dictionary
   */
  public double getExperimentMinimumIDF(int experimentID) {
    PreparedStatement getExpMinimumIDF = null;
    double min = 0.0;
    try {
      getExpMinimumIDF = this.db.prepareStatement("select minimum_idf from experiment where experiment_id = (?)");
      getExpMinimumIDF.setInt(1, experimentID);
      this.results = getExpMinimumIDF.executeQuery();
      if (this.results.next()) {
        min = this.results.getDouble(1);
      }
      getExpMinimumIDF.close();
    } catch (SQLException e) {
      ExceptionHandler.handleSQLException(e, this.db, getExpMinimumIDF);
    }
    return min;
  }

  /**
   * Provides the maximum normalized IDF for an experiment's dictionary
   * 
   * @param experimentID
   *          the <code>int</code> experiment ID to look up in the database
   * 
   * @return the <code>double</code> maximum normalized IDF used in the dictionary
   */
  public double getExperimentMaximumIDF(int experimentID) {
    PreparedStatement getExpMaximumIDF = null;
    double max = 1.0;
    try {
      getExpMaximumIDF = this.db.prepareStatement("select maximum_idf from experiment where experiment_id = (?)");
      getExpMaximumIDF.setInt(1, experimentID);
      this.results = getExpMaximumIDF.executeQuery();
      if (this.results.next()) {
        max = this.results.getDouble(1);
      }
      getExpMaximumIDF.close();
    } catch (SQLException e) {
      ExceptionHandler.handleSQLException(e, this.db, getExpMaximumIDF);
    }
    return max;
  }

  /**
   * Provides the count or percent size of the sample of files used for an experiment
   * 
   * @param experimentID
   *          the <code>int</code> experiment ID to look up in the database
   * 
   * @return the <code>double</code> count or percent of files used; returns 0.0 if sample setting does not exist in
   *         database
   */
  public double getExperimentSampleSetting(int experimentID) {
    PreparedStatement getExpSampleSetting = null;
    double parameter = 0.0;
    try {
      getExpSampleSetting = this.db.prepareStatement("select sample_setting from experiment where experiment_id = (?)");
      getExpSampleSetting.setInt(1, experimentID);
      this.results = getExpSampleSetting.executeQuery();
      if (this.results.next()) {
        parameter = this.results.getDouble(1);
      }
      getExpSampleSetting.close();
    } catch (SQLException e) {
      ExceptionHandler.handleSQLException(e, this.db, getExpSampleSetting);
    }
    return parameter;
  }

  /**
   * Provides the count or percent size of files used for the dictionary in an experiment
   * 
   * @param experimentID
   *          the <code>int</code> experiment ID to look up in the database
   * 
   * @return the <code>double</code> count or percent of files used; returns 0.0 if dictionary setting does not exist in
   *         database
   */
  public double getExperimentDictionarySetting(int experimentID) {
    PreparedStatement getExpDictionarySetting = null;
    double parameter = 0.0;
    try {
      getExpDictionarySetting = this.db
          .prepareStatement("select dictionary_setting from experiment where experiment_id = (?)");
      getExpDictionarySetting.setInt(1, experimentID);
      this.results = getExpDictionarySetting.executeQuery();
      if (this.results.next()) {
        parameter = this.results.getDouble(1);
      }
      getExpDictionarySetting.close();
    } catch (SQLException e) {
      ExceptionHandler.handleSQLException(e, this.db, getExpDictionarySetting);
    }
    return parameter;
  }

  /**
   * Provides the language used in an experiment
   * 
   * @param experimentID
   *          the <code>int</code> experiment ID to look up in the database
   * 
   * @return the string language used; returns an empty string if a language is not present in the database for the
   *         provided experiment
   */
  public String getExperimentLanguage(int experimentID) {
    PreparedStatement getExpLanguage = null;
    String language = "";
    try {
      getExpLanguage = this.db.prepareStatement("select language from experiment where experiment_id = (?)");
      getExpLanguage.setInt(1, experimentID);
      this.results = getExpLanguage.executeQuery();
      if (this.results.next()) {
        language = this.results.getString(1);
      }
      getExpLanguage.close();
    } catch (SQLException e) {
      ExceptionHandler.handleSQLException(e, this.db, getExpLanguage);
    }
    return language;
  }

  /**
   * Provides the mangler names and settings used in an experiment
   * 
   * @param experimentID
   *          the <code>int</code> experiment ID to look up in the database
   * 
   * @return the list of mangler names and settings
   */
  public List<String> getExperimentManglerSettings(int experimentID) {
    List<String> settings = new Vector<String>();
    PreparedStatement getExpManglerSettings = null;
    try {
      getExpManglerSettings = this.db.prepareStatement("select mangler_name, mangler_setting "
          + "from mangler where mangler_id in (select print_mangler_id from experiment_results where experiment_id=?) "
          + "order by mangler_setting");
      getExpManglerSettings.setInt(1, experimentID);
      this.results = getExpManglerSettings.executeQuery();
      while (this.results.next()) {
        settings.add(this.results.getString("mangler_name") + " " + this.results.getInt("mangler_setting"));
      }
      getExpManglerSettings.close();
    } catch (SQLException e) {
      ExceptionHandler.handleSQLException(e, this.db, getExpManglerSettings);
    }
    return settings;
  }

  /**
   * Provides the tokenizer names used in an experiment
   * 
   * @param experimentID
   *          the <code>int</code> experiment ID to look up in the database
   * 
   * @return the list of tokenizer names
   */
  public List<String> getExperimentTokenizers(int experimentID) {
    List<String> tokenizers = new Vector<String>();
    PreparedStatement getExpTokenizers = null;
    try {
      getExpTokenizers = this.db.prepareStatement("select tokenizer.tokenizer_name from experiment_tokenizer "
          + "join tokenizer on tokenizer.tokenizer_id = experiment_tokenizer.tokenizer_id where experiment_id=? "
          + "order by experiment_tokenizer_id;");
      getExpTokenizers.setInt(1, experimentID);
      this.results = getExpTokenizers.executeQuery();
      while (this.results.next()) {
        tokenizers.add(this.results.getString("tokenizer_name"));
      }
      getExpTokenizers.close();
    } catch (SQLException e) {
      ExceptionHandler.handleSQLException(e, this.db, getExpTokenizers);
    }
    return tokenizers;
  }

  /**
   * Provides the number of trials in an experiment
   * 
   * @param experimentID
   *          the <code>int</code> experiment ID to look up in the database
   * 
   * @return the <code>int</code> number of trials
   */
  public int getExperimentTrialCount(int experimentID) {
    PreparedStatement getExpTrialCount = null;
    int count = 0;
    try {
      getExpTrialCount = this.db.prepareStatement("select count(*) from experiment_trial where experiment_id = ?");
      getExpTrialCount.setInt(1, experimentID);
      this.results = getExpTrialCount.executeQuery();
      if (this.results.next()) {
        count = this.results.getInt(1);
      }
      getExpTrialCount.close();
    } catch (SQLException e) {
      ExceptionHandler.handleSQLException(e, this.db, getExpTrialCount);
    }
    return count;
  }
}
