package edu.georgetown.gucs.experiment;

import java.io.File;
import java.io.IOException;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Set;

import edu.georgetown.gucs.dictionary.Dictionary;
import edu.georgetown.gucs.fingerprinter.Fingerprinter;
import edu.georgetown.gucs.matcher.ExactFingerprintMatcher;
import edu.georgetown.gucs.matcher.FingerprintMatcher;
import edu.georgetown.gucs.utility.ExceptionHandler;
import edu.georgetown.gucs.utility.Global;
import edu.georgetown.gucs.utility.ObjectBuilder;
import edu.georgetown.gucs.utility.Pair;

/**
 * Uses a given set of parameters to create a dictionary and fingerprints of a set of documents. Matches are based on
 * comparisons of unmangled and mangled versions of each file and optionally stores the matches in a database.
 * 
 * @author Lindsay Neubauer
 * @author Clay Shields
 * 
 */
public class Trial implements Runnable {

  private TrialParameters parameters;
  private int experimentID;
  private boolean insert;
  private DBInterface database;
  private List<File> sampleFiles;
  private Random random;
  private long randomSeed;
  private String manglers;
  private Fingerprinter fingerprinter;
  private FingerprintMatcher matcher;
  private Dictionary dictionary;
  private int trialID;
  private int dictionaryID;
  private boolean finished = false;

  private int manglerIDCount = 0;
  private int fileIDCount = 0;
  private int fingerprintIDCount = 0;

  private Map<Integer, byte[]> fingerprintID_byte = new HashMap<Integer, byte[]>();
  private Map<byte[], Integer> fingerprintByte_id = new HashMap<byte[], Integer>();

  private Map<Integer, Integer> fingerprintID_manglerID = new HashMap<Integer, Integer>();
  private Map<Integer, Integer> fingerprintID_fileID = new HashMap<Integer, Integer>();

  private Map<Integer, String> manglerID_settings = new HashMap<Integer, String>();
  private Map<String, Integer> manglerSettings_id = new HashMap<String, Integer>();

  private Map<Integer, File> fileID_path = new HashMap<Integer, File>();
  private Map<File, Integer> filePath_id = new HashMap<File, Integer>();

  private Set<Integer> unmangledFingerprints = Collections.synchronizedSet(new HashSet<Integer>());
  private Set<Integer> mangledFingerprints = Collections.synchronizedSet(new HashSet<Integer>());

  private Map<String, Double[]> manglerResults = new LinkedHashMap<String, Double[]>();

  /**
   * Constructor that sets default values for this trial
   * 
   * @param experimentID
   *          the <code>int</code> value of the experiment id this trial is a part of; set to -1 if insert = false
   * @param insert
   *          the <code>boolean</code> value of whether to insert the results from this trial into the database
   * @param database
   *          the dBInterface database to insert trial results; null if insert = false
   * @param dataset
   *          the string name of the database to insert trial results
   * @param sampleFiles
   *          the list of sample files to use in this trial
   * @param randomSeed
   *          the <code>long</code> the initial seed to use for the random number generator in this trial; can be used
   *          to enable repeatability across trials
   * @param dictionaryParameter
   *          the <code>double</code> containing the count or percent to use for dictionary creation in this trial;
   *          numbers less than 1 are processed as percents, numbers greater than or equal to 1 are processed as counts
   * @param sampleParameter
   *          the <code>double</code> containing the count or percent to use for the file sample size in this trial;
   *          numbers less than 1 are processed as percents, numbers greater than or equal to 1 are processed as counts
   * @param dictionary
   *          the dictionary used for this trial
   * @param manglers
   *          the list different mangler strings to use in this trial
   * @param fingerprinter
   *          the string name of the fingerprinter to use in this trial
   * @param matcher
   *          the string name of the matcher to use for fingerprint comparisons in this trial
   * @param matcherScore
   *          the <code>int</code> minimum score that the matcher considers a match
   */
  public Trial(int experimentID, boolean insert, DBInterface database, String dataset, List<File> sampleFiles,
      long randomSeed, double dictionaryParameter, double sampleParameter, Dictionary dictionary, String manglers,
      String fingerprinter, String matcher, int matcherScore) {

    this.experimentID = experimentID;
    this.insert = insert;
    this.database = database;
    this.dictionary = dictionary;
    this.sampleFiles = sampleFiles;
    this.randomSeed = randomSeed;
    this.random = new Random(randomSeed);
    this.fingerprinter = (Fingerprinter) ObjectBuilder.buildObject(fingerprinter);
    this.matcher = (FingerprintMatcher) ObjectBuilder.buildObject(matcher);
    this.matcher.setMinimumScore(matcherScore);
    this.manglers = manglers;

    this.parameters = new TrialParameters(dataset, Double.toString(sampleParameter),
        Double.toString(dictionaryParameter), Double.toString(this.dictionary.getMinIDF()),
        Double.toString(this.dictionary.getMaxIDF()), this.dictionary.getTokenizers().toString(), manglers,
        this.fingerprinter.getClass().getSimpleName(), this.matcher.getClass().getSimpleName(),
        Integer.toString(matcherScore));
  }

  /**
   * Provides the parameters for this trial that are needed for results comparisons
   * 
   * @return the trialParameters containing the relevant parameters
   */
  public TrialParameters getTrialParameters() {
    return this.parameters;
  }

  /**
   * Provides the results for each mangler; sleeps until the trial is finished
   * 
   * @return the map containing the manglers results from this trial
   */
  public Map<String, Double[]> getManglerResults() {
    while (!this.finished) {
      try {
        Thread.sleep(1000);
      } catch (InterruptedException e) {
        ExceptionHandler.handleExceptionExit(e);
      }
    }
    return this.manglerResults;
  }

  /**
   * Runs this trial
   * 
   * @return the <code>int</code> trialID from this trial
   */
  public int runTrial() {
    run();
    return this.trialID;
  }

  /**
   * Runs the trial
   */
  @Override
  public void run() {

    if (this.insert) {
      this.trialID = this.database.newTrial(this.randomSeed);
      this.database.addTrialToExperiment(this.experimentID, this.trialID);
    }

    // Global.monitor.printMessage("Trial", "Creating Dictionary");
    // this.dictionary = new Dictionary();
    // this.dictionary.setLocale(this.language);
    // this.dictionary.setTokenizers(this.tokenizers);
    // this.dictionary.createPartialDictionary(this.datasetFiles);
    // Global.monitor.printMessage("Trial", "Finsihed Partial Dictionary Creation");
    //
    // // trim dictionary if needed before writing
    // if ((this.minIDF > 0) || (this.maxIDF < 1)) {
    // Global.monitor.printMessage("Trial", "Trimming dictionary: min=" + this.minIDF + "  max=" + this.maxIDF);
    // this.dictionary.trimByIDF(this.minIDF, this.maxIDF);
    // }

    if (this.insert) {
      this.dictionaryID = this.database.newDictionary(this.dictionary);
      this.database.addDictionaryToTrial(this.trialID, this.dictionaryID);
    }

    Global.monitor.printMessage("Trial", "Setting up fingerprints");
    this.fingerprinter.setDictionary(this.dictionary);
    this.fingerprinter.setManglerRNG(this.random);
    this.fingerprinter.setMangler(this.manglers, this.dictionary);

    Global.monitor.printMessage("Trial", "Computing null fingerprints for " + this.sampleFiles.size() + " files");
    computeFingerprints(null, this.sampleFiles.iterator());
    Global.monitor.printMessage("Trial", "Computing mangler fingerprints: mangler(s)=" + this.manglers);
    computeFingerprints(this.manglers, this.sampleFiles.iterator());

    Global.monitor.printMessage("Trial", "doComparisons");
    doComparisons();
    Global.monitor.printMessage("Trial", "calculateTrialResults");
    calculateTrialResults();
  }

  /**
   * Computes fingerprints for each of the given files
   * 
   * @param manglerSettings
   *          the string containing the mangler settings to use; null if taking unmangled fingerprints
   * @param files
   *          the file iterator for the files to fingerprint
   */
  private void computeFingerprints(String manglerSettings, Iterator<File> files) {
    String mangler = manglerSettings;
    boolean unmangled = false;
    if (manglerSettings == null) {
      unmangled = true;
      mangler = "null";
    }

    Global.monitor.printMessage("Trial", "Computing fingerprints for  " + manglerSettings);
    int fileCounter = 0;
    int currentManglerID, currentFileID, currentFingerprintID;
    while (files.hasNext()) {
      File currentFile = files.next();
      fileCounter++;

      for (Pair<byte[], String> fingerprintPair : this.fingerprinter.computeFingerprint(currentFile.toString())) {
        // TODO: ! Make sure that the byte information works
        if (fingerprintPair.getB().length() > 0) {
          try {
            currentFile = File.createTempFile(
                currentFile.toString() + "_bytes" + fingerprintPair.getB().replace(" ", "_"), "");
          } catch (IOException e) {
            ExceptionHandler.handleExceptionExit(e);
          }
        }

        if (this.fileID_path.containsValue(currentFile)) {
          currentFileID = this.filePath_id.get(currentFile);
        } else {
          this.fileIDCount++;
          currentFileID = this.fileIDCount;
          this.fileID_path.put(currentFileID, currentFile);
          this.filePath_id.put(currentFile, currentFileID);
        }

        if (this.insert) {
          currentManglerID = this.database.getManglerID(mangler);
          if (this.manglerIDCount < currentManglerID)
            this.manglerIDCount = currentManglerID;
          this.database.addManglerToTrial(this.trialID, currentManglerID);
          this.database.insertFingerprint(currentFile, fingerprintPair.getA(), this.trialID, currentManglerID);
        } else if (this.manglerID_settings.containsValue(mangler)) {
          currentManglerID = this.manglerSettings_id.get(mangler);
          if (this.manglerIDCount < currentManglerID)
            this.manglerIDCount = currentManglerID;
        } else {
          this.manglerIDCount++;
          currentManglerID = this.manglerIDCount;
        }
        this.manglerID_settings.put(currentManglerID, mangler);
        this.manglerSettings_id.put(mangler, currentManglerID);

        if (this.fingerprintID_byte.containsValue(fingerprintPair.getA())) {
          currentFingerprintID = this.fingerprintByte_id.get(fingerprintPair.getA());
        } else {
          this.fingerprintIDCount++;
          currentFingerprintID = this.fingerprintIDCount;
          this.fingerprintID_byte.put(currentFingerprintID, fingerprintPair.getA());
          this.fingerprintID_manglerID.put(currentFingerprintID, currentManglerID);
          this.fingerprintID_fileID.put(currentFingerprintID, currentFileID);
        }

        if (unmangled) {
          this.unmangledFingerprints.add(currentFingerprintID);
        } else {
          this.mangledFingerprints.add(currentFingerprintID);
        }
      }
    }
    Global.monitor.writeToBuffer("Trial", "computeFingerprints", fileCounter + " files");
    Global.monitor.printMessage("Trial", "computeFingerprints: " + fileCounter + " files");
  }

  /**
   * Compares each mangled and unmangled fingerprint using this trials matcher and an exactFingerprintMatcher and
   * records true-positive, false-positive, true-negative, and false-negative results based on this comparison
   */
  private void doComparisons() {
    Global.monitor.writeToBuffer("Trial", "doComparisons", "mangled fingerprints: " + this.mangledFingerprints.size()
        + "  unmangled fingerprint: " + this.unmangledFingerprints.size());
    Global.monitor.printMessage("Trial", "Comparing " + this.mangledFingerprints.size() + " mangled fingerprints to "
        + this.unmangledFingerprints.size() + " unmangled fingerprints");

    ExactFingerprintMatcher exactMatcher = new ExactFingerprintMatcher();
    for (Integer mangledFingerprintID : this.mangledFingerprints) {
      byte[] mangledFP = this.fingerprintID_byte.get(mangledFingerprintID);
      String mangledFile = this.fileID_path.get(this.fingerprintID_fileID.get(mangledFingerprintID)).getPath();

      for (Pair<byte[], String> mangledFPWithoutMangling : this.fingerprinter.computeFingerprint(mangledFile)) {
        for (Integer unmangledFingerprintID : this.unmangledFingerprints) {
          byte[] unmangledFP = this.fingerprintID_byte.get(unmangledFingerprintID);

          // Match and compare to an ExactFingerprint match of the unmangled fingerprints from each file
          boolean matchClaim = this.matcher.match(unmangledFP, mangledFP);
          boolean exactMatch = exactMatcher.match(unmangledFP, mangledFPWithoutMangling.getA());

          // Set up results
          String mangler = this.manglerID_settings.get(this.fingerprintID_manglerID.get(mangledFingerprintID));
          Double[] result = this.manglerResults.get(mangler);
          if (result == null) {
            result = new Double[] { 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0 };
          }

          // Update results
          if (matchClaim == true && exactMatch == true) {
            result[Global.TRUE_POSITIVE]++;
          } else if (matchClaim == true && exactMatch == false) {
            result[Global.FALSE_POSITIVE]++;
          } else if (matchClaim == false && exactMatch == false) {
            result[Global.TRUE_NEGATIVE]++;
          } else if (matchClaim == false && exactMatch == true) {
            result[Global.FALSE_NEGATIVE]++;
          }
          result[Global.DICTIONARY_SIZE] = (double) this.dictionary.size();
          this.manglerResults.put(mangler, result);

          if (this.insert) {
            int score = this.matcher.getScore(unmangledFP, mangledFP);
            this.database.addToFingerprintMatch(this.trialID, unmangledFingerprintID, mangledFingerprintID, matchClaim,
                exactMatch, score);
          }
        }
      }
    }
  }

  /**
   * Computes the precision, recall, and f-score for each mangler
   */
  public void calculateTrialResults() {
    Global.monitor.writeToBuffer("Trial", "calculateTrialResults", "");
    for (Integer fingerprintID : this.mangledFingerprints) {
      String mangler = this.manglerID_settings.get(this.fingerprintID_manglerID.get(fingerprintID));
      Double[] result = this.manglerResults.get(mangler);

      // precision == ppv
      if ((result[Global.TRUE_POSITIVE] + result[Global.FALSE_POSITIVE]) > 0.0) {
        result[Global.PRECISION] = result[Global.TRUE_POSITIVE]
            / (result[Global.TRUE_POSITIVE] + result[Global.FALSE_POSITIVE]);
      } else {
        result[Global.PRECISION] = 0.0;
      }

      // recall == sensitivity
      if ((result[Global.TRUE_POSITIVE] + result[Global.FALSE_NEGATIVE]) > 0.0) {
        result[Global.RECALL] = result[Global.TRUE_POSITIVE]
            / (result[Global.TRUE_POSITIVE] + result[Global.FALSE_NEGATIVE]);
      } else {
        result[Global.RECALL] = 0.0;
      }

      // harmonic mean of precision and recall
      if ((result[Global.PRECISION] + result[Global.RECALL]) > 0) {
        result[Global.FSCORE] = 2 * ((result[Global.PRECISION] * result[Global.RECALL]) / (result[Global.PRECISION] + result[Global.RECALL]));
      } else {
        result[Global.FSCORE] = 0.0;
      }

      if (this.insert) {
        Global.monitor.writeToBuffer("Trial", "calculateTrialResults", "insert into database");
        this.database.addTrialResults(this.trialID, this.fingerprintID_manglerID.get(fingerprintID),
            result[Global.TRUE_POSITIVE].intValue(), result[Global.TRUE_NEGATIVE].intValue(),
            result[Global.FALSE_POSITIVE].intValue(), result[Global.FALSE_NEGATIVE].intValue(),
            result[Global.PRECISION], result[Global.RECALL], result[Global.FSCORE]);
      }
      this.manglerResults.put(mangler, result);
    }

    this.finished = true;
    if (this.insert) {
      Global.monitor.printMessage("Trial", "Trial " + this.trialID + " Finished");
    } else {
      Global.monitor.printMessage("Trial", "Trial Finished");
    }
  }
}