package edu.georgetown.gucs.fingerprinter;

import java.io.File;
import java.io.IOException;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;
import java.util.Random;
import java.util.Vector;

import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;

import edu.georgetown.gucs.utility.AntlrBitSet;
import edu.georgetown.gucs.utility.ExceptionHandler;
import edu.georgetown.gucs.utility.Global;
import edu.georgetown.gucs.utility.Pair;

/**
 * Creates a fingerprint that consist of a bit vector representing the presence or absence of terms in a particular
 * dictionary
 * 
 * @author Clay Shields
 * @author Lindsay Neubauer
 **/
public class BitVectorFingerprinter extends Fingerprinter {

  /**
   * Constructor that generates the fingerprint name, version, unique identifier (GUID), system identifier, and creating
   * program. Sets the fingerprinter output to display all available information.
   */
  public BitVectorFingerprinter() {
    super();
  }

  /**
   * Constructor that loads a dictionary and its tokenizers. Generates the fingerprint name, version, unique identifier
   * (GUID), system identifier, and creating program. Sets the fingerprinter output to display all available
   * information.
   * 
   * @param dictionaryFilename
   *          the filename containing the dictionary to use for this fingerprinter
   **/
  public BitVectorFingerprinter(String dictionaryFilename) {
    super(dictionaryFilename);
  }

  /**
   * Loads a previously generated fingerprinter digest from fingerprinter and dictionary XML files. <br>
   * <br>
   * If dictionaryFilename is null, assumes the full dictionary is contained in the fingerprinter XML file or the file
   * containing the dictionary specified in the fingerprinter XML file exists in the current directory. If the full
   * dictionary is contained in the fingerprinter XML file but does not exist in the current directory, a dictionary XML
   * file containing the dictionary is created.
   * 
   * @param fingerprintFilename
   *          the string name of the XML file containing the fingerprinter's digest
   * @param dictionaryFilename
   *          the filename containing the dictionary to use for this fingerprinter
   */
  public BitVectorFingerprinter(String fingerprintFilename, String dictionaryFilename) {
    super(fingerprintFilename, dictionaryFilename);
  }

  /**
   * Computes a byte array fingerprint indicating the presence or absence of each token in this dictionary
   **/
  @SuppressWarnings("unchecked")
  @Override
  public List<Pair<byte[], String>> computeFingerprint(String filename) {
    if (filename == null) {
      ExceptionHandler.handleExceptionThrow(new Exception("BitVectorFingerprinter: Null filename"));
    }
    if (this.dictionary == null) {
      ExceptionHandler.handleExceptionThrow(new Exception("BitVectorFingerprinter: Null dictionary"));
    }
    if (this.dictionary.getTokenizers() == null) {
      ExceptionHandler.handleExceptionThrow(new Exception("BitVectorFingerprinter: Null tokenizer list"));
    }

    Vector<Pair<byte[], String>> bitVectorList = new Vector<Pair<byte[], String>>();
    AntlrBitSet bitvector = new AntlrBitSet(this.dictionary.size());

    if (containsSplitter(filename)) {
      try {
        SAXBuilder builder = new SAXBuilder();
        Document doc = builder.build(new File(filename));
        Element splitter = doc.getRootElement();
        for (String splitType : Arrays.asList(new String[] { "proportion", "chunks", "w-shingle", "w-shingle-unique" })) {
          for (Element splitterElement : (List<Element>) splitter.getChildren(splitType)) {
            for (Element splitElement : (List<Element>) splitterElement.getChildren("split")) {
              // Clear and resize the bitset
              bitvector = new AntlrBitSet(this.dictionary.size());

              for (Element tokenElement : (List<Element>) splitElement.getChildren("token")) {
                int position = this.dictionary.getPosition(tokenElement.getAttributeValue("name"));
                if (position > -1) {
                  bitvector.add(position);
                }
              }

              bitVectorList.add(new Pair<byte[], String>(bitvector.toBytes(), splitElement
                  .getAttributeValue("start_byte") + " " + splitElement.getAttributeValue("end_byte")));
            }
          }
        }
      } catch (JDOMException e) {
        ExceptionHandler.handleExceptionExit(e);
      } catch (IOException e) {
        ExceptionHandler.handleExceptionExit(e);
      }
    } else {
      // get the tokens from the tokenizer list
      Iterator<String> tokens = this.dictionary.getTokenizers().tokenizeFile(filename);
      while (tokens != null && tokens.hasNext()) {
        int position = this.dictionary.getPosition(tokens.next());
        if (position > -1) {
          bitvector.add(position);
        }
      }
      bitVectorList.add(new Pair<byte[], String>(bitvector.toBytes(), ""));
    }
    return bitVectorList;
  }

  /**
   * Creates a bit vector fingerprint representing the presence or absence of terms in a particular dictionary
   * 
   * @param args
   *          array of string command line arguments<br>
   *          <code>args[0]</code> the filename containing the dictionary to use for this fingerprinter<br>
   *          <code>args[1]</code> the filename of the document to fingerprint<br>
   *          <code>args[2]</code> optional: the filename to write the XML output; if not specified, uses
   *          fingerprint_randomInteger.xml<br>
   *          <code>args[3]</code> optional: the filename containing the output specifications
   */
  public static void main(String args[]) {
    String dictionary_file = "";
    String input_file = "";
    String output_file = "";
    String output_config = null;

    try {
      dictionary_file = args[0];
      input_file = args[1];
      output_file = (args.length > 2) ? args[2] : "fingerprint_" + (new Random()).nextInt(Integer.MAX_VALUE) + ".xml";
      if (args.length > 3) {
        output_config = args[3];
      }
    } catch (ArrayIndexOutOfBoundsException e) {
      System.err
          .println("Usage: java BitVectorFingerprinter <dictionary file> <file to fingeprint> "
              + "<optional: destination filename (otherwise use: fingerprint_<randomInteger>.xml)> <optional: output configuration file>");
      System.exit(-1);
    }

    try {
      BitVectorFingerprinter bvfp = new BitVectorFingerprinter(dictionary_file);
      if (output_config != null) {
        bvfp.outputFields(output_config);
      }
      System.out.println(bvfp.generateXML(bvfp.computeFingerprintXML(input_file), input_file, output_file));

      Global.monitor.closeBuffer();
    } catch (RuntimeException exception) {
      ExceptionHandler.handleExceptionExit(exception);
    }
  }
}
