package edu.georgetown.gucs.fingerprinter;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.UnsupportedEncodingException;
import java.math.BigInteger;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.List;
import java.util.Random;
import java.util.Vector;

import org.apache.xerces.impl.dv.util.Base64;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;
import org.jdom.output.Format;
import org.jdom.output.XMLOutputter;

import edu.georgetown.gucs.dictionary.Dictionary;
import edu.georgetown.gucs.tokenizers.TokenizerList;
import edu.georgetown.gucs.utility.ExceptionHandler;
import edu.georgetown.gucs.utility.Global;
import edu.georgetown.gucs.utility.Pair;

/**
 * This is the base class for all fingerprinters. A Fingerprinter tokenizes a document and compares it to the tokens in
 * a {@linkplain Dictionary} to create a fingerprint that can be used for recognition of the document or comparisons to
 * other documents and fingerprints. <br>
 * For experimental purposes, the fingerprinter controls a <code>mangler</code> which is used to add errors or noise to
 * the files being processed. This allows modification of the input dictionary on the fly, so that fingerprinters can be
 * tested for accuracy over errors.
 * 
 * @author Clay Shields
 **/
public class Fingerprinter {

  /** the dictionary to use to create this fingerprinter */
  protected Dictionary dictionary;
  /** the Base64 encoded string containing the fingerprint */
  protected List<Pair<String, String>> base64Fingerprints = new Vector<Pair<String, String>>();
  /** the name of this fingerprinter */
  protected String fingerprintName;
  /** the globally unique identifier for this fingerprinter */
  protected String GUID;
  /** the fully qualified domain name for the local host IP address */
  protected String systemID;
  /** the name of the program that created the fingerprint */
  protected String creatingProgram;
  /** the person or organization that created this fingerprint */
  protected String creator = "";
  /** the filename of the document this fingerprint is for */
  protected String targetFile;
  /** the version number of this fingerprint */
  protected int version = 1;
  /** the disk image identifier where the fingerprinted document is stored */
  protected String diskImage;
  /** the volume where the fingerprinted document is stored */
  protected String volume;
  /** the byte run for the fingerprinted document */
  protected String byteRun;
  /** the list of unknown tokens contained in this fingerprint */
  protected List<String> unknownTokens;
  /** whether to store digest information in the output for the fingerprint */
  protected boolean showDigest;
  /** whether to store data source information in the output for the fingerprint */
  protected boolean showDataSource;
  /** whether to store the full dictionary used for the fingerprint in the output */
  protected boolean showDictionary;

  /**
   * Constructor that generates the fingerprint name, version, unique identifier (GUID), system identifier, and creating
   * program. Sets the fingerprinter output to display all available information.
   */
  public Fingerprinter() {
    this.fingerprintName = this.getClass().getName();
    this.GUID = new BigInteger(128, new Random()).toString();
    this.creatingProgram = generateCreatingProgram();
    try {
      this.systemID = java.net.InetAddress.getLocalHost().getCanonicalHostName();
    } catch (UnknownHostException e) {
      ExceptionHandler.handleExceptionExit(e);
    }
    setVerbose();
  }

  /**
   * Constructor that loads a dictionary and its tokenizers. Generates the fingerprint name, version, unique identifier
   * (GUID), system identifier, and creating program. Sets the fingerprinter output to display all available
   * information.
   * 
   * @param dict
   *          the dictionary to use to create this fingerprinter
   */
  public Fingerprinter(Dictionary dict) {
    this.dictionary = dict;
    this.fingerprintName = this.getClass().getName();
    this.GUID = new BigInteger(128, new Random()).toString();
    this.creatingProgram = generateCreatingProgram();
    try {
      this.systemID = java.net.InetAddress.getLocalHost().getCanonicalHostName();
    } catch (UnknownHostException e) {
      ExceptionHandler.handleExceptionExit(e);
    }
    setVerbose();
  }

  /**
   * Constructor that loads a dictionary and its tokenizers. Generates the fingerprint name, version, unique identifier
   * (GUID), system identifier, and creating program. Sets the fingerprinter output to display all available
   * information.
   * 
   * @param dictionaryFilename
   *          the filename containing the dictionary to use for this fingerprinter
   **/
  public Fingerprinter(String dictionaryFilename) {
    this.dictionary = new Dictionary(dictionaryFilename);
    this.fingerprintName = this.getClass().getName();
    this.GUID = new BigInteger(128, new Random()).toString();
    this.creatingProgram = generateCreatingProgram();
    try {
      this.systemID = java.net.InetAddress.getLocalHost().getCanonicalHostName();
    } catch (UnknownHostException e) {
      ExceptionHandler.handleExceptionExit(e);
    }

    setVerbose();
  }

  /**
   * Loads a previously generated fingerprinter digest from fingerprinter and dictionary XML files. <br>
   * If dictionaryFilename is null, assumes the full dictionary is contained in the fingerprinter XML file or the file
   * containing the dictionary specified in the fingerprinter XML file exists in the current directory. If the full
   * dictionary is contained in the fingerprinter XML file but does not exist in the current directory, a dictionary XML
   * file containing the dictionary is created.
   * 
   * @param fingerprintFilename
   *          the string name of the XML file containing the fingerprinter's digest
   * @param dictionaryFilename
   *          the name of the XML file containing the dictionary
   **/
  @SuppressWarnings("unchecked")
  public Fingerprinter(String fingerprintFilename, String dictionaryFilename) {
    FileOutputStream digestOut = null;
    try {
      File xml = new File(fingerprintFilename);
      SAXBuilder builder = new SAXBuilder();
      Document doc = builder.build(xml);
      Element fingerprinter = doc.getRootElement();

      this.fingerprintName = fingerprinter.getAttributeValue("fingerprinter");
      this.GUID = fingerprinter.getAttributeValue("GUID");
      this.version = 1 + Integer.valueOf(fingerprinter.getAttributeValue("version"));
      this.systemID = fingerprinter.getAttributeValue("systemID");
      this.creatingProgram = fingerprinter.getAttributeValue("creatingProgram");
      this.creator = fingerprinter.getAttributeValue("creator");

      for (Element fingerprintElement : (List<Element>) fingerprinter.getChildren("base64Fingerprint")) {
        this.base64Fingerprints.add(new Pair<String, String>(fingerprintElement.getTextTrim(), fingerprintElement
            .getAttributeValue("start_byte") + " " + fingerprintElement.getAttributeValue("end_byte")));
      }

      this.showDataSource = !fingerprinter.getChildren("dataSource").isEmpty();
      if (this.showDataSource) {
        Element data = fingerprinter.getChild("dataSource");
        this.targetFile = data.getAttributeValue("filename");
        this.volume = data.getAttributeValue("volume");
        this.diskImage = data.getAttributeValue("diskImageID");
        this.byteRun = data.getAttributeValue("byteRun");
      }

      this.showDigest = !fingerprinter.getChildren("digest").isEmpty();
      if (this.showDigest) {
        Element digest = fingerprinter.getChild("digest");
        if (!digest.getChildren("unknownTokens").isEmpty()) {
          this.unknownTokens = new Vector<String>();
          for (Element unknownToken : (List<Element>) digest.getChild("unknownTokens").getChildren()) {
            this.unknownTokens.add(unknownToken.getAttributeValue("name"));
          }
        }
      }

      Element dict = fingerprinter.getChild("dictionary");
      this.showDictionary = !dict.getChildren().isEmpty();
      String filename = dict.getAttributeValue("name");
      File dictionaryCheck = new File(filename);
      if (dictionaryFilename != null) {
        this.dictionary = new Dictionary(dictionaryFilename);
        this.dictionary.setCreatingProgram(generateCreatingProgram());
      } else if (this.showDictionary && !dictionaryCheck.exists()) {
        Document document = new Document((Element) dict.clone());
        Format form = Format.getPrettyFormat();
        form.setEncoding("UTF-8");
        XMLOutputter out = new XMLOutputter(form);
        digestOut = new FileOutputStream(filename);
        out.output(document, digestOut);
        digestOut.close();
        this.dictionary = new Dictionary(filename);
        TokenizerList tokenizerList = new TokenizerList();
        for (Element tokenizer : (List<Element>) dict.getChildren("tokenizer")) {
          tokenizerList.addTokenizer(tokenizer.getTextTrim());
        }
        this.dictionary.setTokenizers(tokenizerList);
        this.dictionary.setCreatingProgram(generateCreatingProgram());
      } else {
        this.dictionary = new Dictionary(filename);
        this.dictionary.setCreatingProgram(generateCreatingProgram());
      }
    } catch (FileNotFoundException e) {
      ExceptionHandler.handleExceptionExit(e);
    } catch (UnsupportedEncodingException e) {
      ExceptionHandler.handleExceptionExit(e);
    } catch (JDOMException e) {
      ExceptionHandler.handleExceptionExit(e);
    } catch (IOException e) {
      ExceptionHandler.handleExceptionExit(e);
    } finally {
      if (digestOut != null)
        try {
          digestOut.close();
        } catch (IOException e) {
          ExceptionHandler.handleExceptionExit(e);
        }
    }
  }

  /**
   * Specifies which information to display in this fingerprinter's digest XML output
   * 
   * @param showDigest
   *          the boolean value specifying whether to display digest information
   * @param showDataSource
   *          the boolean value specifying whether to display data source information
   * @param showDictionary
   *          the boolean value specifying whether to display dictionary information
   **/
  public void setOutput(boolean showDigest, boolean showDictionary, boolean showDataSource) {
    this.showDigest = showDigest;
    this.showDictionary = showDictionary;
    this.showDataSource = showDataSource;
  }

  /**
   * Sets this fingerprinter's digest XML output to only display file and fingerprint information
   **/
  public void setTerse() {
    setOutput(false, false, false);
  }

  /**
   * Sets this fingerprinter's digest XML output to display all available information
   **/
  public void setVerbose() {
    // TODO: Change to all true once digest information implemented
    setOutput(false, true, true);
  }

  /**
   * Sets the creator of this fingerprinter
   * 
   * @param creator
   *          the person or organization that created this fingerprint; set to an empty string by default
   **/
  public void setCreator(String creator) {
    this.creator = creator;
  }

  /**
   * Passes the specified mangler settings to the mangler for this fingerprinter. The mangler is disabled if "null" is
   * passed as the settings.
   * 
   * @param settings
   *          the string of mangler settings for this fingerprinter
   */
  public void setMangler(String settings) {
    if (settings.contains("null")) {
      this.dictionary.getTokenizers().disableMangler();
    } else {
      this.dictionary.getTokenizers().enableMangler(settings);
    }
  }

  /**
   * Passes the specified mangler settings and a set of tokens to the mangler for this fingerprinter. The mangler is
   * disabled if "null" is passed as the settings.
   * 
   * @param settings
   *          the string of mangler settings for this fingerprinter
   * @param dictionary
   *          the dictionary to use to pass dictionary tokens to the mangler
   */
  public void setMangler(String settings, Dictionary dictionary) {
    if (dictionary == null) {
      Global.monitor.printErrorMessage("Fingerprinter", "Null dictionary passed to setMangler");
      Global.monitor.writeToBuffer("Fingerprinter", "setMangler", "Null dictionary passed to setMangler");
    } else {
      if (settings.contains("null")) {
        this.dictionary.getTokenizers().disableMangler();
      } else {
        this.dictionary.getTokenizers().enableMangler(settings, dictionary.getTokens());
      }
    }
  }

  /**
   * Sets a random number generator for the mangler to allow for repeatability. Repeatable behavior is sometimes used
   * when running experiments.
   * 
   * @param random
   *          the random object to use for this dictionary
   **/
  public void setManglerRNG(Random random) {
    this.dictionary.getTokenizers().setManglerRNG(random);
  }

  /**
   * Gives this fingerprinter's name. This is usually determined by getting the Class name.
   * 
   * @return the string name of this fingerprinter
   **/
  public String getFingerprintName() {
    return this.fingerprintName.substring(this.fingerprintName.lastIndexOf('.') + 1);
  }

  /**
   * Gives this fingerprinter's dictionary
   * 
   * @return the dictionary used for this fingerprinter
   **/
  public Dictionary getDictionary() {
    return this.dictionary;
  }

  /**
   * Gives this fingerprinter's fingerprint in Base64 encoding
   * 
   * @return the Base64 encoded string containing the fingerprint
   **/
  public List<Pair<String, String>> getBase64Fingerprints() {
    return this.base64Fingerprints;
  }

  /**
   * Gives the program that created this fingerprinter
   * 
   * @return the string of the creating program for this fingerprinter
   **/
  public String getCreatingProgram() {
    return this.creatingProgram;
  }

  /**
   * Determines the program that created this fingerprinter
   * 
   * @return the string of the creating program or null if none can be determined
   **/
  public String generateCreatingProgram() {
    this.creatingProgram = "";
    Runtime runtime = Runtime.getRuntime();
    String cmds[] = { "jps", "-m" };
    Process proc;
    try {
      proc = runtime.exec(cmds);
      InputStreamReader in = new InputStreamReader(proc.getInputStream(), "UTF-8");
      BufferedReader bufferedreader = new BufferedReader(in);

      String line;
      while ((line = bufferedreader.readLine()) != null) {
        for (String output : line.split("\\s")) {
          if (output.contains(".jar")) {
            bufferedreader.close();
            in.close();
            return output;
          }
        }
      }
      bufferedreader.close();
      in.close();
    } catch (IOException e) {
      ExceptionHandler.handleExceptionExit(e);
    }
    return null;
  }

  /**
   * Computes the fingerprint of this document as a byte array; indicates the presence or absence of each token in this
   * dictionary
   * 
   * @param filename
   *          the string filename of the document to fingerprint
   * 
   * @return the byte array containing the fingerprint and if applicable the start and end bytes for a split file
   */
  public List<Pair<byte[], String>> computeFingerprint(String filename) {
    return Collections.singletonList(new Pair<byte[], String>(new byte[] {}, ""));
  }

  /**
   * Computes the fingerprint of a document as a Base64 encoded string; indicates the presence or absence of each token
   * in this dictionary
   * 
   * @param filename
   *          the string filename of the document to fingerprint
   * 
   * @return the Base64 encoded string containing the fingerprint and if applicable the start and end bytes for a split
   *         file
   */
  public List<Pair<String, String>> computeFingerprintXML(String filename) {
    ArrayList<Pair<String, String>> base64EncodingList = new ArrayList<Pair<String, String>>();
    for (Pair<byte[], String> byteArrayPair : computeFingerprint(filename)) {
      base64EncodingList.add(new Pair<String, String>(Base64.encode(byteArrayPair.getA()), byteArrayPair.getB()));
    }
    return base64EncodingList;
  }

  /**
   * Computes all the fingerprints of documents in a given directory as Base64 encoded strings; indicates the presence
   * or absence of each token in this dictionary
   * 
   * @param directory
   *          the directory to fingerprint
   * 
   * @return the Base64 encoded string containing the fingerprint and if applicable the start and end bytes for a split
   *         file
   */
  public List<Pair<String, String>> computeAllFingerprintsXML(File directory) {
    ArrayList<Pair<String, String>> base64EncodingList = new ArrayList<Pair<String, String>>();

    if (directory.isDirectory()) {
      for (File filename : directory.listFiles()) {
        if (filename.isFile()) {
          for (Pair<byte[], String> byteArrayPair : computeFingerprint(filename.getAbsolutePath())) {
            base64EncodingList.add(new Pair<String, String>(Base64.encode(byteArrayPair.getA()), byteArrayPair.getB()));
          }
        }
      }
    }
    return base64EncodingList;
  }

  /**
   * Sets the dictionary for this fingerprinter. This dictionary's tokens are compared against the tokenized document
   * that is being fingerprinted.
   * 
   * @param dict
   *          the dictionary to use for this fingerprinter
   **/
  public void setDictionary(Dictionary dict) {
    this.dictionary = dict;
    this.fingerprintName = this.getClass().getName();
  }

  /**
   * Sets the dictionary for this fingerprinter. This dictionary's tokens are compared against the tokenized document
   * that is being fingerprinted.
   * 
   * @param filename
   *          the string filename of the dictionary to use for this fingerprinter
   **/
  public void setDictionary(String filename) {
    this.dictionary = new Dictionary(filename);
    this.fingerprintName = this.getClass().getName();
  }

  /**
   * Checks that the given XML file contains splitter output
   * 
   * @param filename
   *          string XML filename to check
   * 
   * @return the boolean indicating whether the XML file contains splitter output
   */
  static boolean containsSplitter(String filename) {
    try {
      SAXBuilder builder = new SAXBuilder();
      Document doc = builder.build(new File(filename));
      Element root = doc.getRootElement();
      if (root.getName().equalsIgnoreCase("splitter")) {
        if (root.getChild("proportion") != null && root.getChild("proportion").getChild("split") != null) {
          return true;
        }
        if (root.getChild("chunks") != null && root.getChild("chunks").getChild("split") != null) {
          return true;
        }
        if (root.getChild("w-shingle") != null && root.getChild("w-shingle").getChild("split") != null) {
          return true;
        }
        if (root.getChild("w-shingle-unique") != null && root.getChild("w-shingle-unique").getChild("split") != null) {
          return true;
        }
      }
    } catch (JDOMException e) {
      return false;
    } catch (IOException e) {
      return false;
    }
    return false;
  }

  /**
   * Generates an XML file that contains this fingerprinter's digest
   * 
   * @param fingerprintPairList
   *          the list of Pairs containing the Base64 encoded string containing the fingerprint and the start and end
   *          byte information for splitter files
   * @param document
   *          the string filename of the document this fingerprint is for
   * @param outputFile
   *          the string XML filename to write the digest to
   * 
   * @return the string output containing the Base64 encoded fingerprint(s)
   */
  public String generateXML(List<Pair<String, String>> fingerprintPairList, String document, String outputFile) {
     this.base64Fingerprints = fingerprintPairList;
    StringBuffer output = new StringBuffer();
    this.targetFile = document;

    Element fingerprintElement = new Element("fingerprint");
    // Meta-data
    fingerprintElement.setAttribute("fingerprinter", this.getClass().getSimpleName());
    fingerprintElement.setAttribute("GUID", this.GUID);
    fingerprintElement.setAttribute("version", String.valueOf(this.version));
    fingerprintElement.setAttribute("systemID", this.systemID);
    if (this.creator != null && this.creator.length() > 0)
      fingerprintElement.setAttribute("creator", this.creator);
    if (this.creatingProgram != null && this.creatingProgram.length() > 0)
      fingerprintElement.setAttribute("creatingProgram", this.creatingProgram);
    fingerprintElement.setAttribute("creation", String.valueOf(new Date()));

    // Data source
    if (this.showDataSource) {
      Element dataSource = new Element("dataSource");
      dataSource.setAttribute("filename", (new File(this.targetFile)).getAbsolutePath());
      dataSource.setAttribute("filepath", this.dictionary.getSource());
      dataSource.setAttribute("systemID", this.dictionary.getSystemID());
      if (this.volume != null) {
        dataSource.setAttribute("volume", this.volume);
      }
      if (this.diskImage != null) {
        dataSource.setAttribute("diskImageID", this.diskImage);
      }
      if (this.byteRun != null) {
        dataSource.setAttribute("byteRun", this.byteRun);
      }
      fingerprintElement.addContent(dataSource);
    }
    

    // Digest
    if (this.showDigest) {
      Element digest = new Element("digest");
      // File segment digested, by position in token stream
      Element fileSegment = new Element("fileSegment");
      if (fileSegment.getAttributes().size() > 0) {
        digest.addContent(fileSegment);
      }
      // Base64 encoded digest
      Element base64 = new Element("base64Encoding");
      if (base64.getAttributes().size() > 0) {
        digest.addContent(base64);
      }
      // Compression setting for digest
      Element compression = new Element("compression");
      if (compression.getAttributes().size() > 0) {
        digest.addContent(compression);
      }

      this.unknownTokens = new Vector<String>();

      // Should be compressed and base64 encoded (Can be reduced by using
      // Bloom filter of discarded tokens from Dictionary, if included)
      if (this.unknownTokens != null && this.unknownTokens.size() > 0) {
        Element unknown = new Element("unknownTokens");
        for (String unknownToken : this.unknownTokens) {
          Element token = new Element("token");
          token.setAttribute("name", unknownToken);
          unknown.addContent(token);
        }
        digest.addContent(unknown);
      }
      fingerprintElement.addContent(digest);
    }
    
    // Fingerprint(s)
    for (Pair<String, String> fingerprintPair : fingerprintPairList) {
      Element base64FingerprintElement = new Element("base64Fingerprint");
      String start_end = "";
      if (fingerprintPair.getB().length() > 0) {
        String[] start_endBytes = fingerprintPair.getB().split("\\s");
        start_end = String.format(" [%s, %s]", "start_byte: " + start_endBytes[0], "end_byte: " + start_endBytes[1]);
        base64FingerprintElement.setAttribute("start_byte", start_endBytes[0]);
        base64FingerprintElement.setAttribute("end_byte", start_endBytes[1]);
      }
      output.append(String.format("%s (Base64)%s: %n%s%n", getFingerprintName(), start_end, fingerprintPair.getA()));
      base64FingerprintElement.setText(fingerprintPair.getA());
      fingerprintElement.addContent(base64FingerprintElement);
    }

    // Dictionary
    try {
      File xml = new File(this.dictionary.getDictionaryFilename());
      SAXBuilder builder = new SAXBuilder();
      Document doc = builder.build(xml);

      if (this.showDictionary) {
        // tokens are ordered alphabetically
        fingerprintElement.addContent((Element) doc.getRootElement().clone());
      } else {
        Element dict = new Element("dictionary");
        dict.setAttribute("name", this.dictionary.getDictionaryFilename());
        dict.setAttribute("GUID", this.dictionary.getGUID());
        dict.setAttribute("version", String.valueOf(this.dictionary.getVersion()));
        if (this.dictionary.getCreatingProgram() != null && this.dictionary.getCreatingProgram().length() > 0)
          dict.setAttribute("creatingProgram", this.dictionary.getCreatingProgram());
        if (this.dictionary.getCreation() != null && this.dictionary.getCreation().length() > 0)
          dict.setAttribute("creation", this.dictionary.getCreation());
        if (this.dictionary.getSource() != null && this.dictionary.getSource().length() > 0)
          dict.setAttribute("source", this.dictionary.getSource());
        fingerprintElement.addContent(dict);
      }

      Document digestDoc = new Document(fingerprintElement);
      FileOutputStream digest = new FileOutputStream(outputFile);
      Format form = Format.getPrettyFormat();
      form.setEncoding("UTF-8");
      XMLOutputter out = new XMLOutputter(form);
      out.output(digestDoc, digest);
      digest.close();
    } catch (FileNotFoundException e) {
      ExceptionHandler.handleExceptionExit(e);
    } catch (UnsupportedEncodingException e) {
      ExceptionHandler.handleExceptionExit(e);
    } catch (JDOMException e) {
      ExceptionHandler.handleExceptionExit(e);
    } catch (IOException e) {
      ExceptionHandler.handleExceptionExit(e);
    }
    
    System.out.println("Generated digest XML: " + (new File(outputFile)).getAbsolutePath());
    return output.toString();
  }

  /**
   * Sets the digest output to generate for this fingerprinter
   * 
   * @param config_file
   *          the string XML filename containing the output specifications
   **/
  @SuppressWarnings("unchecked")
  public void outputFields(String config_file) {
    setTerse();

    try {
      File xml = new File(config_file);
      SAXBuilder builder = new SAXBuilder();
      Document doc = builder.build(xml);
      Element fingerprinter = doc.getRootElement();

      for (Element output : (List<Element>) fingerprinter.getChildren("output")) {
        if (output.getAttributeValue("type").contains("dataSource"))
          this.showDataSource = true;
        if (output.getAttributeValue("type").contains("digest"))
          this.showDigest = true;
        if (output.getAttributeValue("type").contains("dictionary"))
          this.showDictionary = true;
      }
    } catch (FileNotFoundException e) {
      Global.monitor.printErrorMessage("Fingerprinter", "Output configuration file not found: Setting terse output");
      Global.monitor.writeToBuffer("Fingerprinter", "outputFields",
          "Output configuration file not found: Setting terse output");
    } catch (UnsupportedEncodingException e) {
      ExceptionHandler.handleExceptionExit(e);
    } catch (JDOMException e) {
      ExceptionHandler.handleExceptionExit(e);
    } catch (IOException e) {
      ExceptionHandler.handleExceptionExit(e);
    }
  }
}
