package edu.georgetown.gucs.matcher;

import java.io.File;
import java.io.Serializable;
import java.util.Comparator;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.TreeMap;
import java.util.Vector;

import edu.georgetown.gucs.dictionary.Dictionary;
import edu.georgetown.gucs.fingerprinter.Fingerprinter;
import edu.georgetown.gucs.utility.ExceptionHandler;
import edu.georgetown.gucs.utility.FileLister;
import edu.georgetown.gucs.utility.Global;
import edu.georgetown.gucs.utility.ObjectBuilder;
import edu.georgetown.gucs.utility.Pair;

/**
 * Outputs a list of files and their scores, comparing a given fingerprint to a directory of files. Files are organized
 * by scores with higher scores appearing higher in the list. Providing a minimum score and a dictionary (if the
 * fingerprint does not already contain one) is optional.
 * 
 * @author Lindsay Neubauer
 */
public class CompareFingerprint implements Iterable<Entry<Integer, List<String>>> {

  /** Comparator that orders scores highest to lowest */
  private static class HightoLowScore implements Comparator<Integer>, Serializable {
    private static final long serialVersionUID = 5679507537454070748L;

    public HightoLowScore() {
    }

    @Override
    public int compare(Integer one, Integer two) {
      return two.compareTo(one);
    }
  }

  private Fingerprinter fingerprint;
  private FingerprintMatcher matcher;
  private FileLister fileLister;
  private int minimum_score;
  private Dictionary dictionary;
  private Map<Integer, List<String>> scores;

  /**
   * Constructor that sets the matcher, fingerprint file and directory to use for this comparison
   * 
   * @param match
   *          the string name of the matcher to use for this comparison
   * @param fingerprintFile
   *          the string filename of the fingerprint to compare with this directory
   * @param directory
   *          the string directory path to compare with this fingerprint
   */
  public CompareFingerprint(String match, String fingerprintFile, String directory) {
    this(match, fingerprintFile, directory, 0, null);
  }

  /**
   * Constructor that sets the matcher, fingerprint file, directory and minimum score to use for this comparison
   * 
   * @param match
   *          the string name of the matcher to use for this comparison
   * @param fingerprintFile
   *          the string filename of the fingerprint to compare with this directory
   * @param directory
   *          the string directory path to compare with this fingerprint
   * @param minimumScore
   *          the <code>int</code> minimum score needed to output a file for this comparison
   */
  public CompareFingerprint(String match, String fingerprintFile, String directory, int minimumScore) {
    this(match, fingerprintFile, directory, minimumScore, null);
  }

  /**
   * Constructor that sets the matcher, fingerprint file, directory, minimum score and dictionary to use for this
   * comparison
   * 
   * @param match
   *          the string name of the matcher to use for this comparison
   * @param fingerprintFile
   *          the string filename of the fingerprint to compare with this directory
   * @param directory
   *          the string directory path to compare with this fingerprint
   * @param minimumScore
   *          the <code>int</code> minimum score needed to output a file for this comparison
   * @param dictionary
   *          the string filename of the dictionary to use for this comparison; this dictionary will only be used if the
   *          fingerprint file does not already contain a full dictionary
   */
  public CompareFingerprint(String match, String fingerprintFile, String directory, int minimumScore, String dictionary) {
    this.matcher = (FingerprintMatcher) ObjectBuilder.buildObject(match);
    this.scores = new TreeMap<Integer, List<String>>(new HightoLowScore());
    this.fileLister = new FileLister(directory);

    this.minimum_score = (minimumScore < 0 || minimumScore > 99) ? 0 : minimumScore;

    this.fingerprint = new Fingerprinter(fingerprintFile, dictionary);
    this.dictionary = this.fingerprint.getDictionary();
  }

  /**
   * Compute a fingerprint for each file in this directory and compare it to this fingerprint using this matcher; if the
   * matcher returns boolean values, true values are given a score of 99 and false values are given a score of 0
   */
  public void compareFingerprint() {
    Fingerprinter fingerprinter = (Fingerprinter) ObjectBuilder.buildObject(this.fingerprint.getFingerprintName());
    fingerprinter.setDictionary(this.dictionary);
    String matcherName = this.matcher.getClass().getSimpleName();
    int score;

    for (File file : this.fileLister) {
      String path = file.getPath();
      for (Pair<String, String> fingerprintPair : fingerprinter.computeFingerprintXML(path)) {
        for (Pair<String, String> thisFingerprint : this.fingerprint.getBase64Fingerprints()) {
          if (matcherName.equals("HashFingerprintMatcher") || matcherName.equals("ExactFingerprintMatcher")) {
            score = (this.matcher.match(thisFingerprint.getA(), fingerprintPair.getA())) ? 99 : 0;
          } else {
            score = this.matcher.getScoreXML(thisFingerprint.getA(), fingerprintPair.getA());
          }

          List<String> updatedScores;
          if (this.scores.containsKey(score)) {
            updatedScores = this.scores.get(score);
          } else {
            updatedScores = new Vector<String>();
          }
          String byte_range = (fingerprintPair.getB().length() > 0) ? " (start_byte: "
              + fingerprintPair.getB().split("\\s")[0] + ", end_byte: " + fingerprintPair.getB().split("\\s")[1] + ")"
              : "";
          updatedScores.add(path + byte_range);
          this.scores.put(score, updatedScores);
        }
      }
    }
  }

  /**
   * Compute a fingerprint for each file in this directory and compare it to each other using this matcher; if the
   * matcher returns boolean values, true values are given a score of 99 and false values are given a score of 0
   */
  public void compareFingerprintDirectory() {
    Fingerprinter fingerprinter = (Fingerprinter) ObjectBuilder.buildObject(this.fingerprint.getFingerprintName());
    fingerprinter.setDictionary(this.dictionary);
    String matcherName = this.matcher.getClass().getSimpleName();
    int score;

    for (File file_outer : this.fileLister) {
      for (Pair<String, String> fingerprintPair_outer : fingerprinter.computeFingerprintXML(file_outer.getPath())) {
        String byte_range_outer = (fingerprintPair_outer.getB().length() > 0) ? " (start_byte: "
            + fingerprintPair_outer.getB().split("\\s")[0] + ", end_byte: "
            + fingerprintPair_outer.getB().split("\\s")[1] + ")" : "";

        for (File file_inner : this.fileLister) {
          for (Pair<String, String> fingerprintPair_inner : fingerprinter.computeFingerprintXML(file_inner.getPath())) {
            String byte_range_inner = (fingerprintPair_inner.getB().length() > 0) ? " (start_byte: "
                + fingerprintPair_inner.getB().split("\\s")[0] + ", end_byte: "
                + fingerprintPair_inner.getB().split("\\s")[1] + ")" : "";

            if (matcherName.equals("HashFingerprintMatcher") || matcherName.equals("ExactFingerprintMatcher")) {
              score = (this.matcher.match(fingerprintPair_outer.getA(), fingerprintPair_inner.getA())) ? 99 : 0;
            } else {
              score = this.matcher.getScoreXML(fingerprintPair_outer.getA(), fingerprintPair_inner.getA());
            }

            List<String> updatedScores;
            if (this.scores.containsKey(score)) {
              updatedScores = this.scores.get(score);
            } else {
              updatedScores = new Vector<String>();
            }
            updatedScores.add(String.format("%s%n%s", file_outer + byte_range_outer, file_inner + byte_range_inner));
            this.scores.put(score, updatedScores);
          }
        }
      }
    }
  }

  /**
   * Print all the files and their similarity scores; if a minimum score is provided, only output those files with
   * scores greater than or equal to the minimum
   */
  public void printScores() {
    boolean printedSomething = false;
    for (int i = 99; i >= this.minimum_score; i--) {
      if (this.scores.containsKey(i)) {
        for (String score : this.scores.get(i)) {
          System.out.println(score + ": Score = " + i);
        }
        System.out.println();
        printedSomething = true;
      }
    }
    if (!printedSomething) {
      System.out.println("CompareDictionary: There are no files in the directory with a score higher than "
          + this.minimum_score);
    }
  }

  @Override
  public Iterator<Entry<Integer, List<String>>> iterator() {
    return this.scores.entrySet().iterator();
  }

  /**
   * Outputs a list of files and their scores, comparing a given fingerprint to a directory of files. Files are
   * organized by scores with higher scores appearing higher in the list. Providing a minimum score and a dictionary (if
   * the fingerprint does not already contain one) is optional.
   * 
   * @param args
   *          array of string command line arguments<br>
   *          <code>args[0]</code> the name of the matcher to use for this comparison<br>
   *          <code>args[1]</code> the filename of the fingerprint to compare with this directory<br>
   *          <code>args[2]</code> the directory path to compare with this fingerprint<br>
   *          <code>args[3]</code> the optional integer minimum score needed to output a file for this comparison<br>
   *          <code>args[4]</code> the optional boolean indicating true if n-squared comparison desired, default: false<br>
   *          <code>args[5]</code> the optional filename of the dictionary to use for this comparison; only used if the
   *          fingerprint file does not contain a full dictionary or a link to a dictionary in the file system
   */
  public static void main(String args[]) {
    try {
      String matcher = args[0];
      String fingerprint = args[1];
      String directory = args[2];
      int min_score = (args.length > 3) ? Integer.valueOf(args[3]) : 0;
      boolean n_squared = (args.length > 4) ? Boolean.valueOf(args[4]) : false;
      String dictionary = (args.length > 5) ? args[5] : null;

      CompareFingerprint compare = new CompareFingerprint(matcher, fingerprint, directory, min_score, dictionary);
      if (!n_squared) {
        compare.compareFingerprint();
      } else {
        compare.compareFingerprintDirectory();
      }
      compare.printScores();
      Global.monitor.closeBuffer();
    } catch (ArrayIndexOutOfBoundsException e) {
      System.err
          .println("Usage: java CompareFingerprint <matcher name> <XML fingeprint file> <directory> <optional: minimum score> "
              + "<optional: true/false for n-squared directory comparison> <optional: dictionaryFilename>");
      System.err
          .println("Outputs a list of files and their scores, comparing a given fingerprint to a directory of files");
      System.exit(-1);
    } catch (RuntimeException exception) {
      ExceptionHandler.handleExceptionExit(exception);
    }
  }
}
