package edu.georgetown.gucs.tokenizers;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.Dictionary;

import org.apache.lucene.analysis.TokenStream;
import org.apache.lucene.analysis.cn.ChineseAnalyzer;
import org.apache.lucene.analysis.cn.smart.SmartChineseAnalyzer;
import org.apache.lucene.analysis.tokenattributes.OffsetAttribute;
import org.apache.lucene.analysis.tokenattributes.TermAttribute;
import org.apache.lucene.util.Version;

import edu.georgetown.gucs.utility.ExceptionHandler;
import edu.georgetown.gucs.utility.Global;
import edu.georgetown.gucs.utility.Pair;

/**
 * Splits contents of a Chinese or Chinese-English text file into tokens using Apache Lucene's
 * {@linkplain ChineseAnalyzer} or {@linkplain SmartChineseAnalyzer}. Tokens are used for {@linkplain Dictionary}
 * creation.
 * 
 * @author Lindsay Neubauer
 */
public class ChineseFileTokenizer extends FileTokenizer {

  private static final long serialVersionUID = 3508035370868840628L;

  /**
   * Constructor that sets the token creation mode to split by smart tokenization using probabilistic word segmentation
   * 
   * @see org.apache.lucene.analysis.cn.smart.SmartChineseAnalyzer
   */
  public ChineseFileTokenizer() {
    this.mode = "smart";
  }

  /**
   * Constructor that sets the token creation mode. If set to "individual" the file is split by one Chinese character as
   * one Chinese word; if set to "smart" the file is split by smart tokenization using probabilistic word segmentation.
   * 
   * @param mode
   *          the string mode to split tokens by one Chinese character as one word or by smart tokenization using
   *          probabilistic word segmentation
   */
  public ChineseFileTokenizer(String mode) {
    this.mode = mode.trim();
  }

  /**
   * Sets the token creation mode. If set to "individual" the file is split by one Chinese character as one Chinese
   * word; if set to "smart" the file is split by smart tokenization using probabilistic word segmentation.
   * 
   * @param mode
   *          the string mode to split tokens by one Chinese character as one word or by smart tokenization using
   *          probabilistic word segmentation
   */
  @Override
  public void setMode(String mode) {
    this.mode = mode.trim();
  }

  /**
   * Splits a Chinese or mixed Chinese-English document into tokens based on the token creation mode. If mode is set to
   * "individual" the file is split by one Chinese character as one Chinese word; if set to "smart" the file is split by
   * smart tokenization using probabilistic word segmentation. Assumes the file is UTF-8 encoded and Version.LUCENE_30
   * for "smart" tokenization.
   * 
   * @param filename
   *          the string filename of the document to split into tokens
   * 
   * @see org.apache.lucene.analysis.cn.ChineseAnalyzer
   * @see org.apache.lucene.analysis.cn.smart.SmartChineseAnalyzer
   */
  @Override
  public void tokenize(String filename) {
    if (this.mode.equalsIgnoreCase("individual")) {
      individual_tokenize(filename);
    } else {
      smart_tokenize(filename);
    }
  }

  /**
   * Splits the document into tokens by one Chinese character as one Chinese word. Numeric tokens are removed. English
   * tokens larger than 1 character are included. Assumes the file is UTF-8 encoded.
   * 
   * @param filename
   *          the string filename of the document to split into tokens
   * 
   * @see org.apache.lucene.analysis.cn.ChineseAnalyzer
   */
  private void individual_tokenize(String filename) {

    try {
      BufferedReader buffer_reader = new BufferedReader(new InputStreamReader(new FileInputStream(filename), "UTF-8"));
      ChineseAnalyzer chineseAnalyzer = new ChineseAnalyzer();
      TokenStream tokenStream = chineseAnalyzer.tokenStream(null, buffer_reader);
      TermAttribute termAttribute = tokenStream.addAttribute(TermAttribute.class);
      OffsetAttribute offsetAttribute = tokenStream.addAttribute(OffsetAttribute.class);
      tokenStream.reset();

      synchronized (this.tokenVector) {
        synchronized (this.positions) {
          this.tokenVector.clear();
          this.positions.clear();

          while (tokenStream.incrementToken()) {
            String token = termAttribute.term();
            this.tokenVector.add(token);
            this.positions.add(new Pair<Integer, Integer>(offsetAttribute.startOffset(), offsetAttribute.endOffset()));
          }
        }
      }
      tokenStream.end();
      tokenStream.close();
      buffer_reader.close();
    } catch (IOException e) {
      ExceptionHandler.handleExceptionExit(e);
    }
  }

  /**
   * Splits document into tokens. Uses probabilistic knowledge to find the optimal word segmentation for Simplified
   * Chinese text. The text is first broken into sentences, then each sentence is segmented into words. Assumes the file
   * is UTF-8 encoded and Version.LUCENE_30 for the SmartChineseAnalyzer.
   * 
   * @param filename
   *          the string filename of the document to split into tokens
   * 
   * @see org.apache.lucene.analysis.cn.smart.SmartChineseAnalyzer
   */
  private void smart_tokenize(String filename) {

    try {
      BufferedReader buffer_reader = new BufferedReader(new InputStreamReader(new FileInputStream(filename), "UTF-8"));
      SmartChineseAnalyzer smartChineseAnalyzer = new SmartChineseAnalyzer(Version.LUCENE_30);
      TokenStream tokenStream = smartChineseAnalyzer.tokenStream(null, buffer_reader);
      TermAttribute termAttribute = tokenStream.addAttribute(TermAttribute.class);
      OffsetAttribute offsetAttribute = tokenStream.addAttribute(OffsetAttribute.class);
      tokenStream.reset();

      synchronized (this.tokenVector) {
        synchronized (this.positions) {
          this.tokenVector.clear();
          this.positions.clear();

          while (tokenStream.incrementToken()) {
            String token = termAttribute.term();
            this.tokenVector.add(token);
            this.positions.add(new Pair<Integer, Integer>(offsetAttribute.startOffset(), offsetAttribute.endOffset()));
          }
        }
      }
      tokenStream.end();
      tokenStream.close();
      buffer_reader.close();
    } catch (IOException e) {
      ExceptionHandler.handleExceptionExit(e);
    }
  }

  /**
   * Tokenizes a Chinese text file into single characters and using a probabilistic model and prints both sets of
   * resulting tokens to the the screen
   * 
   * @param args
   *          array of string command line arguments<br>
   *          <code>args[0]</code> the filename of the Chinese language document to split into tokens
   */
  public static void main(String[] args) {
    String input_file = "";
    try {
      input_file = args[0];
    } catch (ArrayIndexOutOfBoundsException e) {
      System.err.println("Usage: java -jar ChineseTokenizer.jar <file name>");
      System.err.println("Tokenizes a Chinese text file into single characters and using a probabilistic model "
          + "and prints both sets of resulting tokens to the the screen");
      System.exit(-1);
    }

    try {
      ChineseFileTokenizer chineseTokenizer = new ChineseFileTokenizer();
      System.out.println("---------------------------");
      System.out.println("Mode: Individual Characters");
      System.out.println("---------------------------");
      chineseTokenizer.setMode("individual");
      chineseTokenizer.tokenize(input_file);
      printTokens(chineseTokenizer.iterator());
      System.out.println("------------------------------------------------");
      System.out.println("Mode: Smart/Probabalistic Character Combinations");
      System.out.println("------------------------------------------------");
      chineseTokenizer.setMode("smart");
      chineseTokenizer.tokenize(input_file);
      printTokens(chineseTokenizer.iterator());
      Global.monitor.closeBuffer();
    } catch (RuntimeException exception) {
      ExceptionHandler.handleExceptionExit(exception);
    }
  }
}
