package edu.georgetown.gucs.tokenizers;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.UnsupportedEncodingException;
import java.util.Dictionary;
import java.util.zip.GZIPInputStream;

import edu.georgetown.gucs.utility.ExceptionHandler;
import edu.georgetown.gucs.utility.Global;
import edu.georgetown.gucs.utility.Pair;

/**
 * Splits contents of a gzipped text file into tokens based on whitespace or by line. Used for {@linkplain Dictionary}
 * creation.
 * 
 * @author Clay Shields
 * @author Lindsay Neubauer
 */
public class GzippedFileTokenizer extends FileTokenizer {

  private static final long serialVersionUID = -1100868800293796306L;

  /**
   * Constructor that sets the token creation mode to split based on whitespace
   */
  public GzippedFileTokenizer() {
    this.mode = "tokens";
  }

  /**
   * Constructor that sets the token creation mode. If set to "lines" the file is split by line; if set to "tokens" the
   * file is split based on whitespace.
   * 
   * @param mode
   *          the string mode to split tokens based on whitespace or by line
   */
  public GzippedFileTokenizer(String mode) {
    this.mode = mode.trim();
  }

  /**
   * Splits a gzipped text document into tokens.
   */
  @Override
  public void tokenize(String filename) {
    if (!filename.contains(".gz")) {
      Global.monitor.printErrorMessage("GzippedFileTokenizer", "Given filename does not contain gzip file extension");
      Global.monitor.writeToBuffer("GzippedFileTokenizer", "tokenize",
          "Given filename does not contain gzip file extension");
    }
    readFile(filename);
  }

  /**
   * Splits a gzipped text document into tokens.
   */
  @Override
  protected void readFile(String filename) {
    BufferedReader buffer_reader = null;
    try {
      // The encoding here needs to match the encoding of the file that is being read; assumes the file is UTF-8 encoded
      // (otherwise, it will spit out garbage)
      buffer_reader = new BufferedReader(new InputStreamReader(new GZIPInputStream(new FileInputStream(filename)),
          "UTF-8"));

      StringBuffer token = new StringBuffer("");
      int character;
      int byte_start = 0;
      int byte_end = 0;

      synchronized (this.tokenVector) {
        synchronized (this.positions) {
          this.tokenVector.clear();
          this.positions.clear();

          while ((character = buffer_reader.read()) != -1) {
            if (this.mode.equalsIgnoreCase("lines")) {
              if ('\n' == character || '\r' == character || '\f' == character) {
                if (token.toString().trim().length() > 0) {
                  this.tokenVector.add(token.toString().trim());
                  this.positions.add(new Pair<Integer, Integer>(byte_start, byte_end));
                  byte_start = byte_end;
                }
                byte_end += Character.charCount(character);
                token.delete(0, token.length());
              } else {
                token.append((char) character);
                byte_end += Character.charCount(character);
              }
            } else {
              if (Character.isWhitespace(character)) {
                if (token.toString().trim().length() > 0) {
                  this.tokenVector.add(token.toString().trim());
                  this.positions.add(new Pair<Integer, Integer>(byte_start, byte_end));
                  byte_start = byte_end;
                }
                byte_end += Character.charCount(character);
                token.delete(0, token.length());
              } else {
                token.append((char) character);
                byte_end += Character.charCount(character);
              }
            }
          }
          buffer_reader.close();
          if (token.toString().trim().length() > 0) {
            this.tokenVector.add(token.toString().trim());
            this.positions.add(new Pair<Integer, Integer>(byte_start, byte_end));
          }
        }
      }
    } catch (UnsupportedEncodingException e) {
      ExceptionHandler.handleExceptionExit(e);
    } catch (FileNotFoundException e) {
      ExceptionHandler.handleExceptionExit(e);
    } catch (IOException e) {
      ExceptionHandler.handleExceptionExit(e);
    } finally {
      try {
        if (buffer_reader != null) {
          buffer_reader.close();
        }
      } catch (IOException e) {
        ExceptionHandler.handleExceptionExit(e);
      }
    }
  }

  /**
   * Prints to the screen a portion of the given gzipped text document
   */
  @Override
  public void checkIndexing(String filename, int start_index, int end_index) {
    try {
      if (!filename.contains(".gz")) {
        Global.monitor.printErrorMessage("GzippedFileTokenizer", "Given filename does not contain gzip file extension");
        Global.monitor.writeToBuffer("GzippedFileTokenizer", "checkIndexing",
            "Given filename does not contain gzip file extension");
      }
      BufferedReader buffer_reader = new BufferedReader(new InputStreamReader(new GZIPInputStream(new FileInputStream(
          filename)), "UTF-8"));
      long byte_position = buffer_reader.skip(start_index);
      StringBuffer str = new StringBuffer("");
      while (byte_position < end_index) {
        int character = buffer_reader.read();
        str.append((char) character);
        byte_position += Character.charCount(character);
      }
      System.out.println(str.toString());
      buffer_reader.close();
    } catch (IOException e) {
      ExceptionHandler.handleExceptionExit(e);
    }
  }

  /**
   * Tokenizes a gzipped text file and prints the resulting tokens to the the screen
   * 
   * @param args
   *          array of string command line arguments<br>
   *          <code>args[0]</code> the filename of the English language gzipped document to split into tokens
   */
  public static void main(String args[]) {
    String filename = "";
    try {
      filename = args[0];
    } catch (ArrayIndexOutOfBoundsException e) {
      System.err.println("Usage: java -jar GzippedFileTokenizer.jar <file name>");
      System.err.println("Tokenizes a gzipped text file and prints the resulting tokens to the the screen");
      System.exit(-1);
    }
    
    if (!filename.endsWith(".gz")) {
      Global.monitor.printErrorMessage("GzippedFileTokenizer", "Given filename does not contain gzip file extension");
      Global.monitor.writeToBuffer("GzippedFileTokenizer", "main",
          "Given filename does not contain gzip file extension");
    } else {
      try {
        GzippedFileTokenizer gzipTokenizer = new GzippedFileTokenizer();
        gzipTokenizer.tokenize(filename);
        printTokens(gzipTokenizer.iterator());
        Global.monitor.closeBuffer();
      } catch (RuntimeException exception) {
        ExceptionHandler.handleExceptionExit(exception);
      }
    }
  }
}
