package edu.georgetown.gucs.tokenizers;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.util.Dictionary;
import edu.georgetown.gucs.utility.ExceptionHandler;

/**
 * Uses Oracle's OutsideIn technology to extract text from a file. Splits contents of the file into tokens based on
 * whitespace or by line. Used for {@linkplain Dictionary} creation.
 * 
 * @author Lindsay Neubauer
 */
public class OutsideInFileTokenizer extends FileTokenizer {

  private static final long serialVersionUID = -7648672098261208008L;

  /**
   * Constructor that sets the token creation mode to split based on whitespace
   */
  public OutsideInFileTokenizer() {
    this.mode = "tokens";
  }

  @Override
  public void tokenize(String filename) {
    callOutsideIn(filename);
  }

  /**
   * Calls Oracle's Outside In API to extract text from a file and splits that text into tokens
   * 
   * @param filename
   *          the string filename of the document to split into tokens
   */
  protected void callOutsideIn(String filename) {
    try {
      File tempFile = File.createTempFile("outsideInOutput", null);
      tempFile.deleteOnExit();
      OutputStreamWriter writer = new OutputStreamWriter(new FileOutputStream(tempFile), "UTF-8");
      InputStream inStream = new ProcessBuilder(
          "/raid/home/neubauer/nps-forensic-triage/code/trunk/lib/outsideInExtract", filename).start().getInputStream();
      BufferedReader buffer = new BufferedReader(new InputStreamReader(inStream, "UTF-8"));
      String line;
      while ((line = buffer.readLine()) != null) {
        writer.write(line);
      }
      writer.close();
      inStream.close();
      buffer.close();

      synchronized (this.tokenVector) {
        synchronized (this.positions) {
          FileTokenizer fileTokenizer = new FileTokenizer(this.mode);
          fileTokenizer.tokenize(tempFile.getAbsolutePath());
          this.tokenVector.clear();
          this.tokenVector.addAll(fileTokenizer.getTokenVector());
          this.positions.clear();
          this.positions.addAll(fileTokenizer.getPositionsVector());
        }
      }
    } catch (FileNotFoundException e) {
      ExceptionHandler.handleExceptionExit(e);
    } catch (IOException e) {
      ExceptionHandler.handleExceptionExit(e);
    }
  }

  /**
   * Tokenizes a file using OutsideIn and prints the resulting tokens to the the screen
   * 
   * @param args
   *          array of string command line arguments<br>
   *          <code>args[0]</code> the filename of the document to split into tokens
   */
  public static void main(String[] args) {
    String filename = "";
    try {
      filename = args[0];
    } catch (ArrayIndexOutOfBoundsException e) {
      System.err.println("Usage: java OutsideInFileTokenizer <file name>");
      System.err.println("Tokenizes a file using OutsideIn and prints the resulting tokens to the the screen");
      System.exit(-1);
    }

    try {
      OutsideInFileTokenizer outIn = new OutsideInFileTokenizer();
      outIn.tokenize(filename);
      printTokens(outIn.iterator());
    } catch (RuntimeException exception) {
      ExceptionHandler.handleExceptionExit(exception);
    }
  }
}
