package edu.georgetown.gucs.tokenizers;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Locale;
import java.util.Map;

import edu.georgetown.gucs.utility.Pair;

/**
 * Eliminates tokens specified in the given stop words document; Assumes this document contains lower case,
 * non-porterized English word
 * 
 * @author Clay Shields
 * @author Lindsay Neubauer
 * 
 */
public class StopWordRemoverTokenizer extends Tokenizer {

  private static final long serialVersionUID = -1767358641887650726L;
  private Map<String, Object> stopwords = new HashMap<String, Object>();

  /**
   * Constructor that loads stop words to use when <code>tokenize</code> method is called;
   * 
   * @param stopwordFilename
   *          the string filename of the document containing the stop words to use; assumes the document contains lower
   *          case, non-porterized English word
   */
  public StopWordRemoverTokenizer(String stopwordFilename) {
    this.stopwords = new HashMap<String, Object>();
    loadStopWords(stopwordFilename);
  }

  /**
   * Loads words to eliminate when <code>tokenize</code> method is called
   * 
   * @param stopwordFilename
   *          the string filename of the document containing the stop words to use; assumes the document contains lower
   *          case, non-porterized English word
   */
  public void loadStopWords(String stopwordFilename) {
    FileTokenizer stops = new FileTokenizer();
    stops.tokenize(stopwordFilename);
    Iterator<String> iterator = stops.iterator();
    while (iterator.hasNext()) {
      String token = iterator.next();
      this.stopwords.put(token, null);
    }
  }

  /**
   * Eliminates tokens specified in the given stop words document
   */
  @Override
  public void tokenize(Iterator<String> iterator) {
    synchronized (this.tokenVector) {
      this.tokenVector.clear();
      while (iterator.hasNext()) {
        String token = iterator.next();
        String lctoken = token.toLowerCase(Locale.ENGLISH);
        if (!this.stopwords.containsKey(lctoken)) {
          this.tokenVector.add(token);
        }
      }
    }
  }

  /**
   * Eliminates tokens specified in the given stop words document
   */
  @Override
  public void tokenize(Iterator<String> tokensIterator, Iterator<Pair<Integer, Integer>> positionsIterator) {
    synchronized (this.tokenVector) {
      synchronized (this.positions) {
        this.tokenVector.clear();
        this.positions.clear();
        while (tokensIterator.hasNext() && positionsIterator.hasNext()) {
          String token = tokensIterator.next().trim();
          Pair<Integer, Integer> pos = positionsIterator.next();
          String lctoken = token.toLowerCase(Locale.ENGLISH);
          if (!this.stopwords.containsKey(lctoken)) {
            this.tokenVector.add(token);
            this.positions.add(pos);
          }
        }
      }
    }
  }
}
