package edu.georgetown.gucs.tokenizers;

import java.io.File;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.Random;
import java.util.Vector;

import edu.georgetown.gucs.utility.ExceptionHandler;
import edu.georgetown.gucs.utility.Global;
import edu.georgetown.gucs.utility.ObjectBuilder;
import edu.georgetown.gucs.utility.Pair;

/**
 * An ordered list of {@linkplain Tokenizer} objects to split a document into tokens and alter the tokens in various
 * ways. The first Tokenizer in this list needs to be able to read directly from a file.
 * 
 * @author Clay Shields
 * @author Lindsay Neubauer
 * 
 */
public class TokenizerList extends Tokenizer {

  private static final long serialVersionUID = -1191618315036978305L;

  // First tokenizer in this list needs to be able to read directly from a file
  private List<Tokenizer> tokenizers;
  private List<String> names;
//  private double minLength = -1;
//  private double maxLength = -1;

  /**
   * Constructor that initializes an empty list of tokenizers
   */
  public TokenizerList() {
    this.tokenizers = new Vector<Tokenizer>();
    this.names = new Vector<String>();
  }

  /**
   * Constructor that takes a list of the tokenizer names
   * 
   * @param tokenizerNames
   *          the string names of the tokenizers to use; the first tokenizer must be able to read from a file
   */
  public TokenizerList(List<String> tokenizerNames) {
    this.tokenizers = new Vector<Tokenizer>();
    this.names = new Vector<String>();
    addTokenizerNames(tokenizerNames);
  }

  /**
   * Writes the non-static and non-transient fields of this class and the string tokenizer names to the output stream
   * 
   * @param out
   *          the objectOutputStream to write this tokenizerList object to
   */
  private void writeObject(ObjectOutputStream out) {
    try {
      out.defaultWriteObject();

      for (String name : this.names) {
        out.writeObject(name);
      }
    } catch (IOException e) {
      ExceptionHandler.handleExceptionExit(e);
    }
  }

  /**
   * Reads the non-static and non-transient fields of this class and the string tokenizer names to the output stream
   * 
   * @param in
   *          the objectInputStream to read this tokenizerList object from
   */
  private void readObject(ObjectInputStream in) {
    try {
      in.defaultReadObject();
      String name;
      while ((name = (String) in.readObject()) != null) {
        this.names.add(name);
      }
    } catch (IOException e) {
      ExceptionHandler.handleExceptionExit(e);
    } catch (ClassNotFoundException e) {
      ExceptionHandler.handleExceptionExit(e);
    }
  }

  /**
   * Enables the mangler with the given settings in the FileManglerTokenizer object in this list
   * 
   * @param settings
   *          the string containing the mangler settings
   */
  public void enableMangler(String settings) {
    FileManglerTokenizer fmt = (FileManglerTokenizer) this.tokenizers.get(findFileMangler());
    fmt.enableMangler(settings);
  }

  /**
   * Enables the mangler with the given settings in the FileManglerTokenizer object in this list
   * 
   * @param settings
   *          the string containing the mangler settings
   * @param tokens
   *          the string list containing the tokens to put through this mangler
   */
  public void enableMangler(String settings, List<String> tokens) {

    if (tokens == null) {
      Global.monitor.printErrorMessage("TokenizerList", "Null list of tokens to mangle");
      Global.monitor.writeToBuffer("TokenizerList", "enableMangler", "Null list of tokens to mangle");
    }

    FileManglerTokenizer fmt = (FileManglerTokenizer) this.tokenizers.get(findFileMangler());
    fmt.enableMangler(settings, tokens);
  }

  /**
   * Disables the mangler in the FileManglerTokenizer object in this list
   */
  public void disableMangler() {
    FileManglerTokenizer fmt = (FileManglerTokenizer) this.tokenizers.get(findFileMangler());
    fmt.disableMangler();
  }

  /**
   * Sets the random number generator to use with a FileManglerTokenizer object in this list
   * 
   * @param random
   *          the random number generator to use if there is a fileManglerTokenizer object in this list
   */
  public void setManglerRNG(Random random) {
    FileManglerTokenizer fmt = (FileManglerTokenizer) this.tokenizers.get(findFileMangler());
    fmt.setRNG(random);
  }

  /**
   * Finds index of FileManglerTokenizer in this list of tokenizers. Assumes there is a FileManglerTokenizer in this
   * list; otherwise an error occurs.
   * 
   * @return the index of the FileManglerTokenizer
   */
  private int findFileMangler() {
    int index = this.names.indexOf("FileManglerTokenizer");
    if (index == -1)
      ExceptionHandler.handleExceptionThrow(new Exception("TokenizerList: FileManglerTokenizer not found"));
    return index;
  }

  /**
   * Adds a tokenizer to the end of this list of tokenizers. Uses reflection to create the correct object. Allows for
   * runtime specification of tokenizer type.
   * 
   * @param tokenizerName
   *          the string name of the tokenizer to be added to the end of this list
   */
  public void addTokenizer(String tokenizerName) {
    Tokenizer toInsert = (Tokenizer) ObjectBuilder.buildObject(tokenizerName);
    this.tokenizers.add(toInsert);
    this.names.add(tokenizerName);
  }

  /**
   * Adds tokenizers to the end of this list of tokenizers
   * 
   * @param tokenizerNames
   *          the list of tokenizers names to use; the first tokenizer must be able to read from a file
   */
  public void addTokenizerNames(List<String> tokenizerNames) {
    for (String tokenizerName : tokenizerNames) {
      addTokenizer(tokenizerName);
    }
  }

  /**
   * Removes the first occurrence of the specified tokenizer name from this tokenizerList
   * 
   * @param tokenizerName
   *          the string name of the tokenizer to remove
   * 
   * @return the <code>boolean</code> value whether this tokenizerList contained the specified tokenizer
   */
  public boolean remove(String tokenizerName) {
    for (int i = 0; i < this.names.size(); i++) {
      if (this.names.get(i).toLowerCase(Locale.ENGLISH).contains(tokenizerName.toLowerCase(Locale.ENGLISH))) {
//        if (tokenizerName.toLowerCase(Locale.ENGLISH).contains("minimumlengthtokenizer")) {
//          // No minimum length tokenizer is set
//          this.minLength = -1;
//        }
//        if (tokenizerName.toLowerCase(Locale.ENGLISH).contains("maximumlengthtokenizer")) {
//          // No maximum length tokenizer is set
//          this.maxLength = -1;
//        }
        this.names.remove(i);
        this.tokenizers.remove(i);
        return true;
      }
    }
    return false;
  }

  /**
   * Clears this list of tokenizers
   **/
  public void clear() {
    this.tokenizers.clear();
    this.names.clear();
  }

  /**
   * Provides the ordered list of the tokenizer names in this list
   * 
   * @return the list of string tokenizer names
   */
  public List<String> getNames() {
    return this.names;
  }

  /**
   * Provides the ordered list of the tokenizers in this list
   * 
   * @return the list of tokenizers
   */
  public List<Tokenizer> getTokenizers() {
    return this.tokenizers;
  }

  @Override
  public String toString() {
    StringBuffer tokenizerNames = new StringBuffer();
    for (String tokenizerName : this.names) {
      tokenizerNames.append(tokenizerName + ", ");
    }
    tokenizerNames.replace(tokenizerNames.lastIndexOf(","), tokenizerNames.length(), "").trimToSize();
    return tokenizerNames.toString();
  }

  /**
   * Applies each tokenizer from this list, in order, on the file; the first tokenizer must be able to read from a file
   * and the tokenizers must already be instantiated
   * 
   * @param filename
   *          the string name of the file to tokenize
   * 
   * @return the string iterator over the tokens remaining after using each tokenizer in this list
   */
  public synchronized Iterator<String> tokenizeFile(String filename) {
    boolean firstTokenizer = true;
    Iterator<String> tokens = null;
    for (Tokenizer tokenizer : this.tokenizers) {
      if (firstTokenizer) {
        tokenizer.tokenize(filename);
        tokens = tokenizer.iterator();
        firstTokenizer = false;
      } else {
        tokenizer.tokenize(tokens);
        tokens = tokenizer.iterator();
      }
    }
    return tokens;
  }

  /**
   * Applies each tokenizer from this list, in order, on the file; the first tokenizer must be able to read from a file
   * and the tokenizers must already be instantiated
   * 
   * @param file
   *          the file to tokenize
   * 
   * @return the string iterator over the tokens remaining after using each tokenizer in this list
   */
  public synchronized Iterator<String> tokenizeFile(File file) {
    return tokenizeFile(file.getAbsolutePath());
  }

  /**
   * Applies each tokenizer from this list, in order, on the tokens provided by a Splitter object; no tokenizers that
   * are able to read from a file should be present in this list
   * 
   * @param tokens_iterator
   *          the string iterator over the token elements
   * @param positions_iterator
   *          the integer pair iterator over the start and end position elements
   * 
   * @return the pair iterator over the tokens and their position pair that remain after using each tokenizer in this
   *         list
   */
  protected Pair<Iterator<String>, Iterator<Pair<Integer, Integer>>> split_tokenize(Iterator<String> tokens_iterator,
      Iterator<Pair<Integer, Integer>> positions_iterator) {
    // iterator over the List of Tokenizer objects
    Iterator<Tokenizer> tokenizerIterator = this.tokenizers.iterator();

    // if there are tokenizers, feed the current tokenVector iterator into the
    // next tokenizer
    Tokenizer current;
    while (tokenizerIterator.hasNext()) {
      current = tokenizerIterator.next();
      current.tokenize(tokens_iterator, positions_iterator);
      tokens_iterator = current.iterator();
      positions_iterator = current.position_iterator();
    }
    // return the remaining tokens and positions after going through each
    // tokenizer in this list
    return new Pair<Iterator<String>, Iterator<Pair<Integer, Integer>>>(tokens_iterator, positions_iterator);
  }
}
