package edu.georgetown.gucs.utility;

import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.UnsupportedEncodingException;
import java.sql.Time;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.zip.GZIPOutputStream;

/**
 * Writes logging information to a GZIP file
 * 
 * @author Lindsay Neubauer
 * 
 */
public class Monitor {

  private OutputStreamWriter writer;
  private File outputFile;
  private long start_time;

  /**
   * Creates a monitor that writes its buffer to a GZIP log file
   */
  public Monitor() {
    this.start_time = System.nanoTime();
    DateFormat dateFormat = new SimpleDateFormat("MMddyyyy_HHmmss_SSSS");
    this.outputFile = new File("log_" + dateFormat.format(new Date(System.currentTimeMillis())) + ".gz");
    openBuffer();
  }

  /**
   * Provides the name of the file where log output is written
   * 
   * @return the string name of the file to output log information
   */
  public String getOutputFile() {
    return this.outputFile.getAbsolutePath();
  }

  /**
   * Opens this monitor's buffer
   */
  private void openBuffer() {
    try {
      this.writer = new OutputStreamWriter(new GZIPOutputStream(new BufferedOutputStream(new FileOutputStream(
          this.outputFile))), "UTF-8");
    } catch (UnsupportedEncodingException e) {
      e.printStackTrace();
    } catch (FileNotFoundException e) {
      e.printStackTrace();
    } catch (IOException e) {
      e.printStackTrace();
    }
  }

  /**
   * Provides the amount of time elapsed since starting this monitor
   * 
   * @return the string representation of the time elapsed since this monitor started running
   */
  private String getElapsedTime() {
    int elapsedSeconds = (int) ((System.nanoTime() - this.start_time) / Math.pow(10, 9));

    int days = elapsedSeconds / 86400;
    int hours = (elapsedSeconds - (days * 86400)) / 3600;
    int minutes = (elapsedSeconds - (days * 86400) - (hours * 3600)) / 60;
    int seconds = elapsedSeconds - (days * 86400) - (hours * 3600) - (minutes * 60);

    return String.format("%02d:%02d:%02d:%02d", days, hours, minutes, seconds);
  }

  /**
   * Prints the status message and the current time for starting this progressMeter
   * 
   * @param classname
   *          the string class name to display for starting activity
   * @param message
   *          the string message to display for starting activity
   */
  public void printStartMessage(String classname, String message) {
    System.out.println("Starting at: " + (new Time(System.currentTimeMillis())).toString());
    printMessage(classname, message);
  }

  /**
   * Prints the status message and time elapsed since starting this progressMeter to System.out
   * 
   * @param classname
   *          the string class name to write to the buffer
   * @param message
   *          the string message to display for this update
   */
  public void printMessage(String classname, String message) {
    System.out.format("%-15s%-35s %s%n", getElapsedTime(), classname, message);
  }

  /**
   * Prints the status message and time elapsed since starting this progressMeter to System.err
   * 
   * @param classname
   *          the string class name to write to the buffer
   * @param message
   *          the string message to display for this update
   */
  public void printErrorMessage(String classname, String message) {
    System.err.format("%-15s%-35s %s%n", getElapsedTime(), classname, message);
  }
  
  /**
   * Prints the status message, time elapsed since starting this progressMeter, and the current time for stopping this
   * progressMeter
   * 
   * @param classname
   *          the string class name to display for ending activity
   * @param message
   *          the string message to display for ending activity
   */
  public void printStopMessage(String classname, String message) {
    printMessage(classname, message);
    System.out.println("Finishing at: " + (new Time(System.currentTimeMillis())).toString());
  }

  /**
   * Adds the given information and the elapsed time to this monitor's buffer
   * 
   * @param classname
   *          the string class name to write to the buffer
   * @param method
   *          the string method name to write to the buffer
   * @param variable
   *          the string variable name and value to write to the buffer
   */
  public void writeToBuffer(String classname, String method, String variable) {
    try {
      this.writer.write(String.format("%-15s%-30s%-30s%-30s%n", getElapsedTime(), classname, method, variable));
    } catch (IOException e) {
      e.printStackTrace();
    }
  }

  /**
   * Adds the given information and the elapsed time to this monitor's buffer
   * 
   * @param info
   *          the string to write to the buffer
   */
  public void writeToBuffer(String info) {
    try {
      this.writer.write(String.format("%-15s%s%n", getElapsedTime(), info));
    } catch (IOException e) {
      e.printStackTrace();
    }
  }

  /**
   * Closes this monitor's buffer, writing it to a GZIP file
   */
  public void closeBuffer() {
    System.out.println();
    System.out.println("Writing log file to " + this.outputFile.getAbsolutePath());

    try {
      writeToBuffer("Monitor", "closeBuffer", "stream closed");
      this.writer.close();
    } catch (IOException e) {
      e.printStackTrace();
    }
  }
}
