package sdtext;

/**
 * Provides the different programs available within SDText (Similarity Digest Text)
 * 
 * @author Lindsay Neubauer
 * 
 */
public enum CommandLineProgramsEnum {

  /**
   * Contains general help information for running SDText via the command line
   */
  HELP("SDText", String.format("%s %n%s %n%s %n%s", 
      "Usage: java -jar sdtext.jar PROGRAM [OPTION]...",
      "Run PROGRAM with specified OPTION(s).", 
      "Example: sdtext TokenizeFile -i document.txt -c tokenizers.xml",
      "Example: sdtext TokenizeFile --input document.txt --config tokenizers.xml"), 
      String.format("%n%s%n %n%s %n%s %n%s %n%s %n%s %n%s %n%s %n%s %n%s %n%s %n%s %n%s %n%s %n%s", 
      "OPTIONS:",
      "-i, --input          filename to read for input", 
      "-c, --config         filename of configuration file",
      "-o, --output         filename to write XML output", 
      "-d, --dictionary     filename containing dictionary",
      "-f, --fingerprint    filename containing fingerprint", 
      "-p, --directory      path to a directory",
      "-m, --matcher        name of matcher to use", 
      "-n, --dataset        name of database to use",
      "-l, --min-idf        minimum normalized IDF (0.0 - 1.0) to consider in dictionary, default = 0.0",
      "-h, --max-idf        maximum normalized IDF (0.0-1.0) to consider in dictionary, default = 1.0",
      "-s, --min-score      minimum score (0-99) needed to output, default = 0", 
      "-b, --bool           boolean value",
      "-a, --action         \"show\" matches on screen or \"insert\" matches into database",
      "-h, --help           display help menu")),

  /**
   * Contains information for running the <code>TokenizeFile</code> program via the command line.
   * <code>TokenizeFile</code> tokenizes a text file and prints the resulting tokens to the the screen.
   */
  TOKENIZEFILE("TokenizeFile", "Tokenizes a text file and prints the resulting tokens to the the screen", String
      .format("%s %n%s", "-i, --input          filename of document to split into tokens",
          "-c, --config         filename of tokenizer configuration file")),

  /**
   * Contains information for running the <code>ArabicFileTokenizer</code> program via the command line.
   * <code>ArabicFileTokenizer</code> tokenizes an Arabic text file and prints the resulting tokens to the the screen.
   */
  ARABICFILETOKENIZER("ArabicFileTokenizer",
      "Tokenizes an Arabic text file and prints the resulting tokens to the the screen",
      "-i, --input          filename of Arabic language document to split into tokens"),

  /**
   * Contains information for running the <code>ChineseFileTokenizer</code> program via the command line.
   * <code>ChineseFileTokenizer</code> tokenizes a Chinese text file into single characters and using a probabilistic
   * model to group characters and prints both sets of resulting tokens to the the screen.
   */
  CHINESEFILETOKENIZER(
      "ChineseFileTokenizer",
      "Tokenizes a Chinese text file into single characters and using a probabilistic model to group characters and prints both sets of resulting tokens to the the screen",
      "-i, --input          filename of Chinese language document to split into tokens"),

  /**
   * Contains information for running the <code>Splitter</code> program via the command line. <code>Splitter</code>
   * tokenizes different chunks of a text file and stores them in an output XML file.
   */
  SPLITTER("Splitter", "Tokenizes different chunks of a text file and stores them in an output XML file", String
      .format("%s %n%s %n%s", "-c, --config         filename of the splitter configuration file",
          "-i, --input          filename of document to split into tokens",
          "-o, --output         filename to write the split chunks of text")),

  /**
   * Contains information for running the <code>CreateDictionary</code> program via the command line.
   * <code>CreateDictionary</code> creates a dictionary from the specified directory using the provided tokenizers.
   */
  CREATEDICTIONARY("CreateDictionary",
      "Creates a dictionary from the specified directory using the provided tokenizers", String.format("%s %n%s %n%s",
          "-c, --config         filename of the tokenizer configuration file",
          "-p, --directory      directory to create the dictionary from",
          "-o, --output         filename to write the dictionary")),

  /**
   * Contains information for running the <code>ShowDictionaryStatistics</code> program via the command line.
   * <code>ShowDictionaryStatistics</code> prints statistics for a dictionary.
   */
  SHOWDICTIONARYSTATISTICS("ShowDictionaryStatistics", "Prints statistics for a dictionary",
      "-d, --dictionary     filename containing the dictionary to show"),

  /**
   * Contains information for running the <code>ShowDictionaryTokens</code> program via the command line.
   * <code>ShowDictionaryTokens</code> prints dictionary tokens and their frequencies, IDFs, and normalized IDFs.
   */
  SHOWDICTIONARYTOKENS("ShowDictionaryTokens",
      "Prints dictionary tokens and their frequencies, IDFs, and normalized IDFs",
      "-d, --dictionary     filename containing the dictionary to show"),

  /**
   * Contains information for running the <code>TrimDictionary</code> program via the command line.
   * <code>TrimDictionary</code> trims a dictionary by removing any token that is outside a range of normalized IDFs.
   */
  TRIMDICTIONARY("TrimDictionary",
      "Trims a dictionary by removing any token that is outside a range of normalized IDFs", String.format(
          "%s %n%s %n%s %n%s", "-d, --dictionary     filename containing the dictionary to trim",
          "-l, --min-idf        minimum normalized IDF (0.0 - 1.0) to keep in trimmed dictionary, default = 0.0",
          "-h, --max-idf        maximum normalized IDF (0.0 - 1.0) to keep in trimmed dictionary, default = 1.0",
          "-o, --output         optional: filename to write the trimmed dictionary")),

  /**
   * Contains information for running the <code>TokenizeFile</code> program via the command line.
   * <code>TokenizeFile</code> tokenizes a text file and prints the resulting tokens to the the screen.
   */
  BITVECTORFINGERPRINTER("BitVectorFingerprinter",
      "Creates a bit vector fingerprint representing the presence or absence of terms in a particular dictionary",
      String.format("%s %n%s %n%s %n%s", "-i, --input          filename of the document to fingerprint",
          "-d, --dictionary     filename containing the dictionary to use",
          "-c, --config         optional: filename containing the output configuration file",
          "-o, --output         optional: filename to write the XML output")),

  /**
   * Contains information for running the <code>ScoreFingerprints</code> program via the command line.
   * <code>ScoreFingerprints</code> compares two XML fingerprints and provides a score for their degree of similarity.
   */
  SCOREFINGERPRINTS("ScoreFingerprints",
      "Compares two XML fingerprints and provides a score for their degree of similarity", String.format(
          "%s %n%s %n%s", "-m, --matcher        matcher name to score the fingerprints",
          "-f, --fingerprint    filename containing fingerprint",
          "-f, --fingerprint    filename containing fingerprint")),

  /**
   * Contains information for running the <code>ExtractFingerprintDictionary</code> program via the command line.
   * <code>ExtractFingerprintDictionary</code> extracts a dictionary from a fingerprint digest and writes it to a new
   * dictionary file.
   */
  EXTRACTFINGERPRINTDICTIONARY("ExtractFingerprintDictionary",
      "Extracts a dictionary from a fingerprint digest and writes it to a new dictionary file", String.format(
          "%s %n%s", "-i, --input          filename containing the fingerprint and its dictionary to extract",
          "-o, --output         filename to write the extracted dictionary")),

  /**
   * Contains information for running the <code>CompareFingerprint</code> program via the command line.
   * <code>CompareFingerprint</code> outputs a list of files and their scores, comparing a given fingerprint to a
   * directory of files.
   */
  COMPAREFINGERPRINT("CompareFingerprint",
      "Outputs a list of files and their scores, comparing a given fingerprint to a directory of files", 
      String.format(
          "%s %n%s %n%s %n%s %n%s %n%s", 
          "-m, --matcher        matcher name to make comparison",
          "-f, --fingerprint    filename containing fingerprint", 
          "-p, --directory      directory path to compare",
          "-s, --min-score      optional: minimum score (0-99) needed to output a file, default = 0",
          "-b, --bool           optional: boolean value (true/false) indicating whether to do an n-squared comparison",
          "-d, --dictionary     optional: filename containing the dictionary to use; only needed if the fingerprint "
              + "file does not contain a full dictionary or a link to a dictionary in the file system")),

  /**
   * Contains information for running the <code>Experiment</code> program via the command line. <code>Experiment</code>
   * outputs results from multiple runs with different experiment settings and optionally stores them in a database.
   */
  EXPERIMENT("Experiment",
      "Outputs results from multiple runs with different experiment settings and optionally stores them in a database",
      "-c, --config         filename of the experiment configuration file");

  final String name;
  final String use;
  final String options;

  /**
   * Stores information about each program
   * 
   * @param name
   *          the string name of the program
   * @param use
   *          the string explaining the purpose of the program
   * @param options
   *          the string specifying the command line options for the program
   */
  CommandLineProgramsEnum(String name, String use, String options) {
    this.name = name;
    this.use = use;
    this.options = options;
  }
}
