package sdtext;

import java.awt.Component;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.Arrays;
import java.util.Iterator;
import java.util.Vector;

import javax.swing.GroupLayout;
import javax.swing.JButton;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.LayoutStyle;
import javax.swing.ScrollPaneConstants;
import javax.swing.SwingConstants;
import javax.swing.SwingWorker;
import javax.swing.table.DefaultTableCellRenderer;
import javax.swing.table.DefaultTableModel;

import edu.georgetown.gucs.tokenizers.TokenizeFile;
import edu.georgetown.gucs.utility.Global;

/**
 * Provides an interface for receiving action events related to a tokenizer; used with {@linkplain GUI}.
 * 
 * @author Lindsay Neubauer
 * 
 */
class TokenizeFileActionListener implements ActionListener {

  TokenizerSelectionHandler tokenizerHandler;
  JButton fileLink = new JButton("");
  JTable tokenizeFileOutputTable;
  DefaultTableModel tableModel;
  JButton fileSelectionButton;
  DefaultTableCellRenderer rightToLeftRenderer;
  DefaultTableCellRenderer leftToRightRenderer;

  /**
   * Constructor that specifies the tokenizer selection handler
   * 
   * @param tokenizerHandler
   *          the handler for tokenizer selection
   */
  public TokenizeFileActionListener(TokenizerSelectionHandler tokenizerHandler) {
    this.tokenizerHandler = tokenizerHandler;
    GUI.linkSetup(this.fileLink);
    this.fileSelectionButton = new JButton("Select File");
    this.fileSelectionButton.addActionListener(this);

    this.tableModel = new DefaultTableModel(new Vector<String>(),
        new Vector<String>(Arrays.asList(new String[] { "" })));
    this.tokenizeFileOutputTable = new JTable(this.tableModel);
    this.tokenizeFileOutputTable.setEnabled(false);
    this.tokenizeFileOutputTable.setFont(Global.unifontFont14);

    this.rightToLeftRenderer = new DefaultTableCellRenderer();
    this.rightToLeftRenderer.setHorizontalAlignment(SwingConstants.RIGHT);
    this.leftToRightRenderer = new DefaultTableCellRenderer();
    this.leftToRightRenderer.setHorizontalAlignment(SwingConstants.LEFT);
  }

  @Override
  public void actionPerformed(ActionEvent actionEvent) {
    this.tableModel.setRowCount(0);
    if (this.tokenizerHandler.getTokenizers().get(0).equalsIgnoreCase("ArabicFileTokenizer")) {
      this.tokenizeFileOutputTable.getColumn("").setCellRenderer(this.rightToLeftRenderer);
    } else {
      this.tokenizeFileOutputTable.getColumn("").setCellRenderer(this.leftToRightRenderer);
    }
    JFileChooser fileChooser = new JFileChooser();
    fileChooser.setDialogTitle("Choose a file to tokenize");
    this.fileLink.setText("<HTML>Tokenizing File...<HTML>");
    this.fileSelectionButton.setEnabled(false);
    int returnVal = fileChooser.showOpenDialog((Component) actionEvent.getSource());
    if (returnVal == JFileChooser.APPROVE_OPTION) {
      try {
        final String filename = fileChooser.getSelectedFile().getAbsolutePath();
        String fileTokenizer = this.tokenizerHandler.getTokenizers().get(0);
        int retVal = 0;
        if (fileTokenizer.contains("ArabicFileTokenizer") || fileTokenizer.contains("ChineseFileTokenizer")) {
          retVal = JOptionPane.showConfirmDialog(this.fileSelectionButton,
              fileTokenizer.split("\\s")[0].replace("FileTokenizer", "") + " language document: " + filename,
              "Language Check: " + fileTokenizer, JOptionPane.YES_NO_OPTION, JOptionPane.ERROR_MESSAGE);
        }

        if (retVal == 0) {
          SwingWorker<String, String> worker = new SwingWorker<String, String>() {
            @Override
            protected String doInBackground() {
              TokenizeFile tokenizeFile = new TokenizeFile();
              tokenizeFile.addTokenizers(TokenizeFileActionListener.this.tokenizerHandler.getTokenizers());
              tokenizeFile.tokenize(filename);
              Iterator<String> tokensIterator = tokenizeFile.iterator();
              while (tokensIterator.hasNext()) {
                TokenizeFileActionListener.this.tableModel.addRow(new String[] { tokensIterator.next() });
              }
              GUI.createLink(filename, TokenizeFileActionListener.this.fileLink, false);
              TokenizeFileActionListener.this.fileSelectionButton.setEnabled(true);
              return "";
            }
          };
          worker.execute();
        } else {
          this.fileSelectionButton.setEnabled(true);
          this.tableModel.setRowCount(0);
          this.fileLink.setText("");
        }
      } catch (RuntimeException exception) {
        this.fileSelectionButton.setEnabled(true);
        this.tableModel.setRowCount(0);
        this.fileLink.setText("");
        Global.monitor.printMessage("TokenizeFileActionListener", "Exception: " + exception.getMessage());
        Global.monitor.writeToBuffer("TokenizeFileActionListener", "fileSelectionButton.actionPerformed",
            exception.getMessage());
        JOptionPane.showConfirmDialog((Component) actionEvent.getSource(), exception.getMessage(), "Error: Matcher",
            JOptionPane.CANCEL_OPTION, JOptionPane.ERROR_MESSAGE);
      }
    } else {
      this.fileSelectionButton.setEnabled(true);
      this.tableModel.setRowCount(0);
      this.fileLink.setText("");
    }
  }

  /**
   * Creates a panel for selecting and tokenizing a file
   * 
   * @return the panel containing the components necessary for selecting and tokenizing a file
   */
  protected JPanel createPanel() {
    JPanel tokenizerPanel = new JPanel();
    GroupLayout layout = GUI.createGroupLayout(tokenizerPanel);

    JLabel titleLabel = new JLabel("Tokenize File");
    titleLabel.setFont(Global.boldUnifontFont);

    JScrollPane outputScrollPane = new JScrollPane(this.tokenizeFileOutputTable,
        ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED, ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED);
    outputScrollPane.setPreferredSize(this.tokenizeFileOutputTable.getPreferredSize());

    layout.setHorizontalGroup(layout.createParallelGroup().addComponent(titleLabel).addComponent(this.fileLink)
        .addComponent(this.fileSelectionButton).addComponent(outputScrollPane));
    layout.setVerticalGroup(layout.createSequentialGroup().addComponent(titleLabel).addComponent(this.fileLink)
        .addComponent(this.fileSelectionButton)
        .addPreferredGap(titleLabel, this.fileLink, LayoutStyle.ComponentPlacement.INDENT)
        .addPreferredGap(titleLabel, this.fileSelectionButton, LayoutStyle.ComponentPlacement.INDENT)
        .addComponent(outputScrollPane));

    return tokenizerPanel;
  }
}
