package sdtext;

/**
 * Provides the different tokenizers available within SDText (Similarity Digest Text)
 * 
 * @author Lindsay Neubauer
 * 
 */
public enum TokenizerEnum {
  
  /**
   * Splits contents of a text file into tokens based on whitespace or by line
   */
  FileTokenizer("FileTokenizer", "tokens lines"),

  /**
   * Splits contents of a gzipped text file into tokens based on whitespace or by line 
   */
  GzippedFileTokenizer("GzippedFileTokenizer", "tokens lines"), 
  
  /**
   * Uses Oracle's OutsideIn technology to extract text from a file
   */
  OutsideInFileTokenizer("OutsideInFileTokenizer", ""), 
      
  /**
   * Splits contents of an Arabic text file into tokens using Apache Lucene's ArabicAnalyzer
   */
  ArabicFileTokenizer("ArabicFileTokenizer", ""), 
  
  /**
   * Splits contents of a Chinese or Chinese-English text file into tokens using Apache Lucene's ChineseAnalyzer or SmartChineseAnalyzer
   */
  ChineseFileTokenizer("ChineseFileTokenizer", "smart individual"), 
  
  /**
   * Eliminates tokens related to attachments in the Enron data set
   */
  EnronEmailTokenizer("EnronEmailTokenizer", "FileTokenizer GzippedFileTokenizer OutsideInFileTokenizer"), 
  
  /**
   * Eliminates tokens from headers in the Enron data set
   */
  EnronStripMailHeaderTokenizer("EnronStripMailHeaderTokenizer", "FileTokenizer GzippedFileTokenizer OutsideInFileTokenizer"), 
  
  /**
   * Eliminates tokens that are longer than a given length
   */
  MaximumLengthTokenizer("MaximumLengthTokenizer", "FileTokenizer GzippedFileTokenizer OutsideInFileTokenizer ArabicFileTokenizer ChineseFileTokenizer"), 
  
  /**
   * Eliminates tokens that are shorter than a given length
   */
  MinimumLengthTokenizer("MinimumLengthTokenizer", "FileTokenizer GzippedFileTokenizer OutsideInFileTokenizer ArabicFileTokenizer ChineseFileTokenizer"), 
  
  /**
   * Uses English stemming to change tokens into their root form
   */
  PorterTokenizer("PorterTokenizer", "FileTokenizer GzippedFileTokenizer OutsideInFileTokenizer"), 
  
  /**
   * Eliminates tokens that are numbers
   */
  RemoveNumericTokensTokenizer("RemoveNumericTokensTokenizer", "FileTokenizer GzippedFileTokenizer OutsideInFileTokenizer ChineseFileTokenizer"), 
  
  /**
   * Eliminates tokens containing numbers
   */
  RemoveTokensWithNumbersTokenizer("RemoveTokensWithNumbersTokenizer", "FileTokenizer GzippedFileTokenizer OutsideInFileTokenizer"), 
  
  /**
   * Eliminates tokens specified in the given stop words document; Assumes this document contains lower case, non-porterized English word
   */
  StopWordRemoverTokenizer("StopWordRemoverTokenizer", "FileTokenizer GzippedFileTokenizer OutsideInFileTokenizer ArabicFileTokenizer ChineseFileTokenizer"), 
  
  /**
   * Eliminates tokens nested inside mark-up language tags; assumes that tokens have been split by line rather than using whitespace
   */
  StripMarkupTokenizer("StripMarkupTokenizer", "FileTokenizer GzippedFileTokenizer"), 
  
  /**
   * Separates tokens based on punctuation and removes punctuation from tokens
   */
  StripPunctuationTokenizer("StripPunctuationTokenizer", "FileTokenizer GzippedFileTokenizer OutsideInFileTokenizer");

  final String name;
  final String parameters;

  /**
   * Stores information about each tokenizer
   * 
   * @param name
   *          the string name of the tokenizer
   * @param parameters
   *          the string array of parameters for the tokenizer
   */
  TokenizerEnum(String name, String parameters) {
    this.name = name;
    this.parameters = parameters;
  }
}
